# This file is automatically generated by pyo3_stub_gen
# ruff: noqa: E501, F401

import builtins
from . import expression
from . import instructions
from . import program
from . import waveforms

class PickleError(builtins.QuilError):
    r"""
    Errors when trying to pickle or deepcopy.
    """
    ...

class QuilError(builtins.Exception):
    r"""
    Base exception type for errors raised by this package.
    """
    ...

class QuilValueError(builtins.QuilError):
    r"""
    Input value is invalid.
    """
    ...

class ToQuilError(builtins.QuilError):
    r"""
    Errors which can occur when converting a Quil item to a string.
    """
    ...




import pytest
from quil.expression import Expression
from quil.instructions import (
    MeasureCalibrationDefinition,
    MeasureCalibrationIdentifier,
    PauliGate,
    PauliTerm,
    Qubit,
    CalibrationDefinition,
    CalibrationIdentifier,
    QubitPlaceholder,
    Instruction,
    Delay,
    Gate,
)

class TestPauliTerm:
    def test_new(self):
        pt = PauliTerm([(PauliGate.X, "a")], Expression.Pi())
        assert pt.arguments == [(PauliGate.X, "a")]
        pt = PauliTerm([(PauliGate.Y, "b")], Expression.Pi())
        assert pt.arguments == [(PauliGate.Y, "b")]


def test_calibration_getters():
    placeholder = Qubit.Placeholder(QubitPlaceholder())

    calibration = CalibrationDefinition(
        CalibrationIdentifier("MYCAL", [], [placeholder], []),
        [Instruction.Delay(Delay(Expression.Number(complex(0.5)), [], [placeholder]))],
    )

    assert calibration.name == "MYCAL"
    assert calibration.parameters == []
    assert calibration.qubits == [placeholder]
    assert calibration.modifiers == []


def test_measure_calibration_getters():
    placeholder = Qubit.Placeholder(QubitPlaceholder())

    calibration = MeasureCalibrationDefinition(
        MeasureCalibrationIdentifier(placeholder, "addr"),
        [],
    )

    assert calibration.qubit == placeholder
    assert calibration.target == "addr"
    assert calibration.instructions == []


@pytest.mark.xfail(reason="This underflows, but should raise a proper exception.")
def test_to_unitary():
    g = Gate("X", (), (Qubit.Fixed(0),), ())
    g.to_unitary(0)

if __name__ == "__main__":
    test_to_unitary()




import pickle
from typing import Optional

import pytest
from syrupy.assertion import SnapshotAssertion

from quil.instructions import Gate, Instruction, Jump, Qubit, QubitPlaceholder, Target, TargetPlaceholder
from quil.program import MaybeCalibrationExpansion, Program
from quil import QuilError


def test_pickle():
    program = Program.parse(
        """
DECLARE ro BIT[2]
H 0
CNOT 0 1
MEASURE 0 ro[0]
MEASURE 1 ro[1]
"""
    )
    pickled = pickle.dumps(program)
    unpickled = pickle.loads(pickled)
    assert program == unpickled


def test_custom_resolver():
    qubit_placeholder = QubitPlaceholder()

    def qubit_resolver(qubit: QubitPlaceholder) -> Optional[int]:
        return {qubit_placeholder: 9}.get(qubit)

    target_placeholder = TargetPlaceholder("base-target")

    def target_resolver(target: TargetPlaceholder) -> Optional[str]:
        print("resolving target", target)
        return {target_placeholder: "test"}.get(target)

    program = Program()
    program.add_instructions(
        [
            Instruction.Gate(Gate("H", [], [Qubit.Placeholder(qubit_placeholder)], [])),
            Instruction.Jump(Jump(Target.Placeholder(target_placeholder))),
        ]
    )

    with pytest.raises(QuilError):
        program.to_quil()

    program.resolve_placeholders_with_custom_resolvers(target_resolver=target_resolver, qubit_resolver=qubit_resolver)

    print(program.to_quil_or_debug())

    assert program.to_quil() == "H 9\nJUMP @test\n"


def test_single_block_control_flow_analysis():
    program = Program.parse(
        """
DECLARE ro BIT[2]
H 0
CNOT 0 1
MEASURE 0 ro[0]
MEASURE 1 ro[1]
"""
    )
    assert program.used_qubits == {Qubit.Fixed(0), Qubit.Fixed(1)}
    cfg = program.control_flow_graph()
    blocks = cfg.basic_blocks()
    assert len(blocks) == 1
    assert blocks[0].terminator is None
    block_program = Program()
    block_program.add_instructions(blocks[0].instructions)
    assert (
        block_program.to_quil()
        == """H 0
CNOT 0 1
MEASURE 0 ro[0]
MEASURE 1 ro[1]
"""
    )


def test_multi_block_control_flow_analysis():
    program = Program.parse(
        """
DECLARE ro BIT[1]
LABEL @start
MEASURE 0 ro[0]
JUMP-UNLESS @end ro[0]
X 0
JUMP @start
LABEL @end
HALT
"""
    )
    assert program.used_qubits == {Qubit.Fixed(0)}
    cfg = program.control_flow_graph()

    assert cfg.has_dynamic_control_flow()

    blocks = cfg.basic_blocks()
    assert len(blocks) == 3


def test_basic_block_fixed_schedule():
    """Test that a simple, realistic program can be scheduled as expected."""
    program = Program.parse(
        """DEFFRAME 0 "flux_tx_cz":
    TEST: 1

DEFFRAME 1 "flux_tx_iswap":
    TEST: 1

DEFFRAME 1 "flux_tx_cz":
    TEST: 1

DEFFRAME 1 "flux_tx_iswap":
    TEST: 1

DEFFRAME 2 "flux_tx_cz":
    TEST: 1

DEFFRAME 2 "flux_tx_iswap":
    TEST: 1

DEFFRAME 3 "flux_tx_cz":
    TEST: 1

DEFFRAME 3 "flux_tx_iswap":
    TEST: 1

# Simplified version
DEFCAL CZ q0 q1:
    FENCE q0 q1
    SET-PHASE q0 "flux_tx_cz" 0.0
    SET-PHASE q1 "flux_tx_iswap" 0.0
    NONBLOCKING PULSE q0 "flux_tx_cz" erf_square(duration: 6.000000000000001e-08, pad_left: 0.1e-08, pad_right: 0.1e-08)
    NONBLOCKING PULSE q1 "flux_tx_iswap" erf_square(duration: 6.000000000000001e-08, pad_left: 0.1e-08, pad_right: 0.1e-08)
    SHIFT-PHASE q0 "flux_tx_cz" 1.0
    SHIFT-PHASE q1 "flux_tx_iswap" 1.0
    FENCE q0 q1

CZ 0 1
CZ 2 3
CZ 0 2
"""
    )
    cfg = program.control_flow_graph()
    blocks = cfg.basic_blocks()
    assert len(blocks) == 1
    block = blocks[0]

    schedule = block.as_schedule_seconds(program)
    items = schedule.items

    # One for each CZ
    assert len(items) == 3

    items = {item.instruction_index: item for item in items}

    # The first CZ should start at 0
    assert items[0].time_span.start == 0.0

    # The first CZ should start and end at the same times
    assert items[0].time_span.start == items[1].time_span.start
    assert items[0].time_span.duration == items[1].time_span.duration

    # The third CZ should start when the first and second ones end and be of the same duration
    assert items[0].time_span.start + items[0].time_span.duration == items[2].time_span.start
    assert items[0].time_span.duration == items[2].time_span.duration


def test_filter_instructions(snapshot: SnapshotAssertion):
    input = """DECLARE foo REAL[1]
DEFFRAME 1 "rx":
\tHARDWARE-OBJECT: "hardware"
DEFCAL I 1:
\tDELAY 0 1
DEFGATE BAR AS MATRIX:
\t0, 1
\t1, 0

H 1
CNOT 2 3
"""
    program = Program.parse(input)
    program_without_quil_t = program.filter_instructions(lambda instruction: not instruction.is_quil_t())
    assert program_without_quil_t.to_quil() == snapshot


def test_calibration_expansion():
    """
    Assert that program calibration expansion happens as expected and that the source map is correct.
    """
    import inspect

    program_text = inspect.cleandoc(
        """
        DEFCAL X 0:
            Y 0

        DEFCAL Y 0:
            Z 0

        X 0
        Y 0
        """
    )
    program = Program.parse(program_text)
    expansion = program.expand_calibrations_with_source_map()
    source_map = expansion.source_map

    expected_program_text = inspect.cleandoc(
        """
        DEFCAL X 0:
            Y 0

        DEFCAL Y 0:
            Z 0

        Z 0
        Z 0
        """
    )

    assert expansion.program.to_quil() == Program.parse(expected_program_text).to_quil()

    # The X at index 0 should have been replaced with a Z at index 0
    targets = source_map.list_targets_for_source_index(0)
    assert len(targets) == 1
    assert isinstance(targets[0], MaybeCalibrationExpansion.Expanded)
    expanded = targets[0][0]
    assert expanded.range == range(0, 1)
    assert source_map.list_sources_for_target_index(0) == [0]

    # The Y at index 1 should have been replaced with a Z at index 1
    targets = source_map.list_targets_for_source_index(1)
    assert len(targets) == 1
    assert isinstance(targets[0], MaybeCalibrationExpansion.Expanded)
    expanded = targets[0][0]
    assert expanded.range == range(1, 2)
    assert source_map.list_sources_for_target_index(1) == [1]

    # There is no source index 2 and so there should be no mapping
    assert source_map.list_targets_for_source_index(2) == []



# This file is automatically generated by pyo3_stub_gen
# ruff: noqa: E501, F401

import builtins

class BoxcarKernel:
    r"""
    A Boxcar waveform.
    """
    @property
    def phase(self) -> builtins.float:
        r"""
        The phase, in cycles.
        """
    @property
    def sample_count(self) -> builtins.int:
        r"""
        Sample count, which must be positive.
        """
    @property
    def scale(self) -> builtins.float:
        r"""
        Scale applied to the waveform envelope.
        """
    @phase.setter
    def phase(self, value: builtins.float) -> None:
        r"""
        The phase, in cycles.
        """
    @sample_count.setter
    def sample_count(self, value: builtins.int) -> None:
        r"""
        Sample count, which must be positive.
        """
    @scale.setter
    def scale(self, value: builtins.float) -> None:
        r"""
        Scale applied to the waveform envelope.
        """
    def __eq__(self, other:builtins.object) -> builtins.bool: ...
    def __new__(cls, phase:builtins.float, scale:builtins.float, sample_count:builtins.int) -> BoxcarKernel:
        r"""
        Create a new `BoxcarKernel`.
        
        This raises a `ValueError` if `sample_count` is zero.
        """
    def __repr__(self) -> builtins.str: ...
    def into_iq_value(self) -> builtins.complex:
        r"""
        Get the `complex` value this `BoxcarKernel` represents.
        """

class DragGaussian:
    r"""
    Creates a waveform with a DRAG-corrected Gaussian shape.
    
    This is a Gaussian shape with an additional component proportional to the time derivative of the main Gaussian pulse.
    
    See Motzoi F. et al., Phys. Rev. Lett., 103 (2009) 110501. for details.
    """
    @property
    def alpha(self) -> builtins.float:
        r"""
        DRAG parameter - controls strength of the imaginary term
        """
    @property
    def anh(self) -> builtins.float:
        r"""
        Qubit anharmonicity - sets rate of evolution for the imaginary term (Hz)
        """
    @property
    def detuning(self) -> builtins.float:
        r"""
        Explicit detuning to bake into iq values
        """
    @property
    def duration(self) -> builtins.float:
        r"""
        Full duration of the pulse (s)
        """
    @property
    def fwhm(self) -> builtins.float:
        r"""
        Full width half maximum of the pulse (s)
        """
    @property
    def phase(self) -> builtins.float:
        r"""
        Phase shift for entire waveform
        """
    @property
    def sample_rate(self) -> builtins.float:
        r"""
        Generate samples at this rate (Hz)
        """
    @property
    def scale(self) -> builtins.float:
        r"""
        Scale to apply to waveform envelope
        """
    @property
    def t0(self) -> builtins.float:
        r"""
        Center/offset for pulse centroid (s)
        """
    @alpha.setter
    def alpha(self, value: builtins.float) -> None:
        r"""
        DRAG parameter - controls strength of the imaginary term
        """
    @anh.setter
    def anh(self, value: builtins.float) -> None:
        r"""
        Qubit anharmonicity - sets rate of evolution for the imaginary term (Hz)
        """
    @detuning.setter
    def detuning(self, value: builtins.float) -> None:
        r"""
        Explicit detuning to bake into iq values
        """
    @duration.setter
    def duration(self, value: builtins.float) -> None:
        r"""
        Full duration of the pulse (s)
        """
    @fwhm.setter
    def fwhm(self, value: builtins.float) -> None:
        r"""
        Full width half maximum of the pulse (s)
        """
    @phase.setter
    def phase(self, value: builtins.float) -> None:
        r"""
        Phase shift for entire waveform
        """
    @sample_rate.setter
    def sample_rate(self, value: builtins.float) -> None:
        r"""
        Generate samples at this rate (Hz)
        """
    @scale.setter
    def scale(self, value: builtins.float) -> None:
        r"""
        Scale to apply to waveform envelope
        """
    @t0.setter
    def t0(self, value: builtins.float) -> None:
        r"""
        Center/offset for pulse centroid (s)
        """
    def __eq__(self, other:builtins.object) -> builtins.bool: ...
    def __new__(cls, duration:builtins.float, fwhm:builtins.float, t0:builtins.float, anh:builtins.float, alpha:builtins.float, sample_rate:builtins.float, scale:builtins.float, phase:builtins.float, detuning:builtins.float) -> DragGaussian: ...
    def __repr__(self) -> builtins.str: ...
    def into_iq_value(self) -> builtins.list[builtins.complex]:
        r"""
        Get the a list of `complex` values from this waveform.
        """

class ErfSquare:
    r"""
    A waveform with a flat top and edges that are error functions (erfs).
    """
    @property
    def detuning(self) -> builtins.float:
        r"""
        Explicit detuning to bake into iq values
        """
    @property
    def duration(self) -> builtins.float:
        r"""
        Full duration of the pulse (s)
        """
    @property
    def pad_left(self) -> builtins.float:
        r"""
        Length of zero padding to add to beginning of pulse (s)
        """
    @property
    def pad_right(self) -> builtins.float:
        r"""
        Length of zero padding to add to end of pulse (s)
        """
    @property
    def phase(self) -> builtins.float:
        r"""
        Phase shift for entire waveform
        """
    @property
    def positive_polarity(self) -> builtins.bool:
        r"""
        Toggle for positive/negative polarity
        """
    @property
    def risetime(self) -> builtins.float:
        r"""
        Slope of erf shoulders (2x FWHM of erf in s)
        """
    @property
    def sample_rate(self) -> builtins.float:
        r"""
        Generate wavform samples at this rate (Hz)
        """
    @property
    def scale(self) -> builtins.float:
        r"""
        Scale to apply to waveform envelope
        """
    @detuning.setter
    def detuning(self, value: builtins.float) -> None:
        r"""
        Explicit detuning to bake into iq values
        """
    @duration.setter
    def duration(self, value: builtins.float) -> None:
        r"""
        Full duration of the pulse (s)
        """
    @pad_left.setter
    def pad_left(self, value: builtins.float) -> None:
        r"""
        Length of zero padding to add to beginning of pulse (s)
        """
    @pad_right.setter
    def pad_right(self, value: builtins.float) -> None:
        r"""
        Length of zero padding to add to end of pulse (s)
        """
    @phase.setter
    def phase(self, value: builtins.float) -> None:
        r"""
        Phase shift for entire waveform
        """
    @positive_polarity.setter
    def positive_polarity(self, value: builtins.bool) -> None:
        r"""
        Toggle for positive/negative polarity
        """
    @risetime.setter
    def risetime(self, value: builtins.float) -> None:
        r"""
        Slope of erf shoulders (2x FWHM of erf in s)
        """
    @sample_rate.setter
    def sample_rate(self, value: builtins.float) -> None:
        r"""
        Generate wavform samples at this rate (Hz)
        """
    @scale.setter
    def scale(self, value: builtins.float) -> None:
        r"""
        Scale to apply to waveform envelope
        """
    def __eq__(self, other:builtins.object) -> builtins.bool: ...
    def __new__(cls, duration:builtins.float, risetime:builtins.float, sample_rate:builtins.float, pad_left:builtins.float, pad_right:builtins.float, positive_polarity:builtins.bool, scale:builtins.float, phase:builtins.float, detuning:builtins.float) -> ErfSquare: ...
    def __repr__(self) -> builtins.str: ...
    def into_iq_value(self) -> builtins.list[builtins.complex]:
        r"""
        Get the a list of `complex` values from this waveform.
        """

class Gaussian:
    r"""
    Creates a waveform with a Gaussian shape.
    """
    @property
    def detuning(self) -> builtins.float:
        r"""
        Explicit detuning to bake into IQ values
        """
    @property
    def duration(self) -> builtins.float:
        r"""
        Full duration of the pulse (s)
        """
    @property
    def fwhm(self) -> builtins.float:
        r"""
        Full width half maximum of the pulse (s)
        """
    @property
    def phase(self) -> builtins.float:
        r"""
        Phase shift for entire waveform
        """
    @property
    def sample_rate(self) -> builtins.float:
        r"""
        Generate waveform samples at this rate (Hz)
        """
    @property
    def scale(self) -> builtins.float:
        r"""
        Scale to apply to waveform envelope
        """
    @property
    def t0(self) -> builtins.float:
        r"""
        Center/offset for pulse centroid (s)
        """
    @detuning.setter
    def detuning(self, value: builtins.float) -> None:
        r"""
        Explicit detuning to bake into IQ values
        """
    @duration.setter
    def duration(self, value: builtins.float) -> None:
        r"""
        Full duration of the pulse (s)
        """
    @fwhm.setter
    def fwhm(self, value: builtins.float) -> None:
        r"""
        Full width half maximum of the pulse (s)
        """
    @phase.setter
    def phase(self, value: builtins.float) -> None:
        r"""
        Phase shift for entire waveform
        """
    @sample_rate.setter
    def sample_rate(self, value: builtins.float) -> None:
        r"""
        Generate waveform samples at this rate (Hz)
        """
    @scale.setter
    def scale(self, value: builtins.float) -> None:
        r"""
        Scale to apply to waveform envelope
        """
    @t0.setter
    def t0(self, value: builtins.float) -> None:
        r"""
        Center/offset for pulse centroid (s)
        """
    def __eq__(self, other:builtins.object) -> builtins.bool: ...
    def __new__(cls, duration:builtins.float, fwhm:builtins.float, t0:builtins.float, sample_rate:builtins.float, scale:builtins.float, phase:builtins.float, detuning:builtins.float) -> Gaussian: ...
    def __repr__(self) -> builtins.str: ...
    def into_iq_value(self) -> builtins.list[builtins.complex]:
        r"""
        Get the a list of `complex` values from this waveform.
        """

class HermiteGaussian:
    r"""
    Creates a Hermite Gaussian waveform.
    
    This extends the basic DRAG pulse by adding an additional imaginary term to the pulse envelope consisting of a
    Gaussian pulse modified by the second order Hermite polynomial.
    
    Refer to
    "Effects of arbitrary laser or NMR pulse shapes on population inversion and coherence"
    Warren S. Warren. 81, (1984); doi: 10.1063/1.447644
    for details.
    """
    @property
    def alpha(self) -> builtins.float:
        r"""
        DRAG parameter - controls strength of the imaginary term
        """
    @property
    def anh(self) -> builtins.float:
        r"""
        Qubit anharmonicity - sets rate of evolution for the imaginary term
        """
    @property
    def detuning(self) -> builtins.float:
        r"""
        Explicit detuning to bake into iq values
        """
    @property
    def duration(self) -> builtins.float:
        r"""
        Full duration of the pulse
        """
    @property
    def fwhm(self) -> builtins.float:
        r"""
        Full width half maximum of the pulse
        """
    @property
    def phase(self) -> builtins.float:
        r"""
        Phase shift for entire waveform
        """
    @property
    def sample_rate(self) -> builtins.float:
        r"""
        Generate samples at this rate
        """
    @property
    def scale(self) -> builtins.float:
        r"""
        Scale to apply to waveform envelope
        """
    @property
    def second_order_hrm_coeff(self) -> builtins.float:
        r"""
        Coefficient of the second order Hermite polynomial term.
        """
    @property
    def t0(self) -> builtins.float:
        r"""
        Center/offset for pulse centroid
        """
    @alpha.setter
    def alpha(self, value: builtins.float) -> None:
        r"""
        DRAG parameter - controls strength of the imaginary term
        """
    @anh.setter
    def anh(self, value: builtins.float) -> None:
        r"""
        Qubit anharmonicity - sets rate of evolution for the imaginary term
        """
    @detuning.setter
    def detuning(self, value: builtins.float) -> None:
        r"""
        Explicit detuning to bake into iq values
        """
    @duration.setter
    def duration(self, value: builtins.float) -> None:
        r"""
        Full duration of the pulse
        """
    @fwhm.setter
    def fwhm(self, value: builtins.float) -> None:
        r"""
        Full width half maximum of the pulse
        """
    @phase.setter
    def phase(self, value: builtins.float) -> None:
        r"""
        Phase shift for entire waveform
        """
    @sample_rate.setter
    def sample_rate(self, value: builtins.float) -> None:
        r"""
        Generate samples at this rate
        """
    @scale.setter
    def scale(self, value: builtins.float) -> None:
        r"""
        Scale to apply to waveform envelope
        """
    @second_order_hrm_coeff.setter
    def second_order_hrm_coeff(self, value: builtins.float) -> None:
        r"""
        Coefficient of the second order Hermite polynomial term.
        """
    @t0.setter
    def t0(self, value: builtins.float) -> None:
        r"""
        Center/offset for pulse centroid
        """
    def __eq__(self, other:builtins.object) -> builtins.bool: ...
    def __new__(cls, duration:builtins.float, fwhm:builtins.float, t0:builtins.float, anh:builtins.float, alpha:builtins.float, sample_rate:builtins.float, second_order_hrm_coeff:builtins.float, scale:builtins.float, phase:builtins.float, detuning:builtins.float) -> HermiteGaussian: ...
    def __repr__(self) -> builtins.str: ...
    def into_iq_value(self) -> builtins.list[builtins.complex]:
        r"""
        Get the a list of `complex` values from this waveform.
        """

def apply_phase_and_detuning(iq_values:list, phase:builtins.float, detuning:builtins.float, sample_rate:builtins.float) -> None:
    r"""
    Modulate and phase shift waveform IQ data in place.
    """




# This file is automatically generated by pyo3_stub_gen
# ruff: noqa: E501, F401

import builtins
import collections.abc
import numpy
import numpy.typing
import quil.instructions
import typing

class BasicBlock:
    @property
    def instructions(self) -> builtins.list[Instruction]:
        r"""
        A list of the instructions in the block, in order of definition.
        
        This does not include the label or terminator instructions.
        """
    @property
    def label(self) -> typing.Optional[Target]:
        r"""
        The label of the block, if any.
        This is used to target this block in control flow.
        """
    @property
    def terminator(self) -> typing.Optional[Instruction]:
        r"""
        The control flow terminator instruction of the block, if any.
        
        If this is ``None``, the implicit behavior is to "continue" to the subsequent block.
        """
    def __new__(cls, instance:BasicBlock) -> BasicBlock: ...
    def __repr__(self) -> builtins.str: ...
    def as_schedule_seconds(self, program:Program) -> ScheduleSeconds: ...
    def gate_depth(self, gate_minimum_qubit_count:builtins.int) -> builtins.int:
        r"""
        Return the length of the longest path
        from an initial instruction (one with no prerequisite instructions)
        to a final instruction (one with no dependent instructions),
        where the length of a path is the number of gate instructions in the path.
        
        :param gate_minimum_qubit_count:
            The minimum number of qubits in a gate
            for it to be counted in the depth.
        """

class BasicBlockScheduleError(builtins.ProgramError):
    ...

class CalibrationExpansion:
    r"""
    Details about the expansion of a calibration.
    """
    @property
    def calibration_used(self) -> CalibrationSource:
        r"""
        The calibration used to expand the instruction.
        """
    @property
    def expansions(self) -> CalibrationExpansionSourceMap:
        r"""
        The source map describing the nested expansions made.
        """
    @property
    def range(self) -> range:
        r"""
        The range of instructions in the expanded list
        which were generated by this expansion.
        """
    def __eq__(self, other:builtins.object) -> builtins.bool: ...
    def __repr__(self) -> builtins.str: ...

class CalibrationExpansionSourceMap:
    def __repr__(self) -> builtins.str: ...
    def entries(self) -> builtins.list[CalibrationExpansionSourceMapEntry]: ...
    def list_sources_for_calibration_used(self, calibration_used:CalibrationSource) -> builtins.list[builtins.int]:
        r"""
        Given a particular calibration (`DEFCAL` or `DEFCAL MEASURE`), =
        return the locations in the source which were expanded using that calibration.
        
        This is `O(n)` where `n` is the number of first-level calibration expansions performed.
        """
    def list_sources_for_target_index(self, target_index:builtins.int) -> builtins.list[builtins.int]:
        r"""
        Return all source ranges in the source map
        which were used to generate the target index.
        
        This is `O(n)` where `n` is the number of entries in the map.
        """
    def list_targets_for_source_index(self, source_index:builtins.int) -> builtins.list[CalibrationExpansion]:
        r"""
        Return all target ranges which were used to generate the source range.
        
        This is `O(n)` where `n` is the number of entries in the map.
        """

class CalibrationExpansionSourceMapEntry:
    r"""
    A description of the expansion of one instruction into other instructions.
    
    If present, the instruction located at `source_location` was expanded using calibrations
    into the instructions located at `target_location`.
    
    Note that both `source_location` and `target_location` are relative to the scope of expansion.
    
    In the case of a nested expansion, both describe the location relative only to that
    level of expansion and *not* the original program.
    
    Consider the following example:
    
    ```python
    DEFCAL A:
        NOP
        B
        HALT
    
    
    DEFCAL B:
        NOP
        WAIT
    
    NOP
    NOP
    NOP
    A
    ```
    
    In this program, `A` will expand into `NOP`, `B`, and `HALT`.
    Then, `B` will expand into `NOP` and `WAIT`.
    Each level of this expansion
    will have its own ``CalibrationExpansionSourceMap`` describing the expansion.
    In the map of `B` to `NOP` and `WAIT`, the `source_location` will be `1`
    because `B` is the second instruction in `DEFCAL A`,
    even though `A` is the 4th instruction (index = 3) in the original program.
    """
    def __repr__(self) -> builtins.str: ...
    def source_location(self) -> builtins.int:
        r"""
        The instruction index within the source program's body instructions.
        """
    def target_location(self) -> CalibrationExpansion:
        r"""
        The location of the expanded instruction within the target
        program's body instructions.
        """

class CalibrationSet:
    r"""
    A collection of Quil calibrations (`DEFCAL` instructions) with utility methods.
    
    This exposes the semantics similar to [`CalibrationSet`] to Python users,
    so see the documentation there for more information.
    """
    @property
    def calibrations(self) -> builtins.list[CalibrationDefinition]:
        r"""
        Return a list of all [`CalibrationDefinition`]s in the set.
        """
    @property
    def measure_calibrations(self) -> builtins.list[MeasureCalibrationDefinition]:
        r"""
        Return a list of all [`MeasureCalibrationDefinition`]s in the set.
        """
    def __eq__(self, other:builtins.object) -> builtins.bool: ...
    def __len__(self) -> builtins.int:
        r"""
        Return the count of contained calibrations.
        """
    def __new__(cls, calibrations:typing.Sequence[CalibrationDefinition], measure_calibrations:typing.Sequence[MeasureCalibrationDefinition]) -> CalibrationSet: ...
    def __repr__(self) -> builtins.str: ...
    def expand(self, instruction:Instruction, previous_calibrations:typing.Sequence[Instruction]) -> typing.Optional[builtins.list[Instruction]]:
        r"""
        Given an instruction, return the instructions to which it is expanded if there is a match.
        Recursively calibrate instructions, returning an error if a calibration directly or indirectly
        expands into itself.
        
        Return only the expanded instructions; for more information about the expansion process,
        see [`Self::expand_with_detail`].
        """
    def extend(self, other:CalibrationSet) -> None:
        r"""
        Append another [`CalibrationSet`] onto this one.
        
        Calibrations with conflicting [`CalibrationSignature`]s are overwritten by the ones in the
        given set.
        """
    def get_match_for_gate(self, gate:Gate) -> typing.Optional[CalibrationDefinition]:
        r"""
        Return the final calibration which matches the gate per the QuilT specification:
        
        A calibration matches a gate if:
        1. It has the same name
        2. It has the same modifiers
        3. It has the same qubit count (any mix of fixed & variable)
        4. It has the same parameter count (both specified and unspecified)
        5. All fixed qubits in the calibration definition match those in the gate
        6. All specified parameters in the calibration definition match those in the gate
        """
    def get_match_for_measurement(self, measurement:Measurement) -> typing.Optional[MeasureCalibrationDefinition]:
        r"""
        Returns the last-specified ``MeasureCalibrationDefinition`` that matches the target
        qubit (if any), or otherwise the last-specified one that specified no qubit.
        
        If multiple calibrations match the measurement, the precedence is as follows:
        
          1. Match fixed qubit.
          2. Match variable qubit.
          3. Match no qubit.
        
        In the case of multiple calibrations with equal precedence, the last one wins.
        """
    def insert_calibration(self, calibration:CalibrationDefinition) -> typing.Optional[CalibrationDefinition]:
        r"""
        Insert a [`CalibrationDefinition`] into the set.
        
        If a calibration with the same [signature][crate::instruction::CalibrationSignature] already
        exists in the set, it will be replaced and the old calibration will be returned.
        """
    def insert_measurement_calibration(self, calibration:MeasureCalibrationDefinition) -> typing.Optional[MeasureCalibrationDefinition]:
        r"""
        Insert a [`MeasureCalibrationDefinition`] into the set.
        
        If a calibration with the same [signature][crate::instruction::CalibrationSignature] already
        exists in the set, it will be replaced and the old calibration will be returned.
        """
    def is_empty(self) -> builtins.bool:
        r"""
        Return true if this contains no data.
        """
    def to_instructions(self) -> builtins.list[Instruction]:
        r"""
        Return the Quil instructions which describe the contained calibrations.
        """

class CalibrationSource:
    r"""
    The source of a calibration, either a [`CalibrationIdentifier`] or a
    [`MeasureCalibrationIdentifier`].
    """
    def __getnewargs__(self) -> tuple[CalibrationIdentifier | MeasureCalibrationIdentifier]: ...
    def __repr__(self) -> builtins.str: ...
    class Calibration(CalibrationSource):
        r"""
        Describes a `DEFCAL` instruction
        """
        __match_args__ = ("_0",)
        @property
        def _0(self) -> CalibrationIdentifier: ...
        def __getitem__(self, key:builtins.int) -> typing.Any: ...
        def __len__(self) -> builtins.int: ...
        def __new__(cls, _0:CalibrationIdentifier) -> CalibrationSource.Calibration: ...
    
    class MeasureCalibration(CalibrationSource):
        r"""
        Describes a `DEFCAL MEASURE` instruction
        """
        __match_args__ = ("_0",)
        @property
        def _0(self) -> MeasureCalibrationIdentifier: ...
        def __getitem__(self, key:builtins.int) -> typing.Any: ...
        def __len__(self) -> builtins.int: ...
        def __new__(cls, _0:MeasureCalibrationIdentifier) -> CalibrationSource.MeasureCalibration: ...
    

class ComputedScheduleError(builtins.ProgramError):
    r"""
    Error raised if the computed schedule is invalid.
    """
    ...

class ControlFlowGraph:
    def __new__(cls, instance:ControlFlowGraph) -> ControlFlowGraph: ...
    def __repr__(self) -> builtins.str: ...
    def basic_blocks(self) -> builtins.list[BasicBlock]:
        r"""
        Return a list of all the basic blocks in the control flow graph, in order of definition.
        """
    def has_dynamic_control_flow(self) -> builtins.bool:
        r"""
        Return ``True`` if the program has dynamic control flow, i.e. contains a conditional branch instruction.
        
        ``False`` does not imply that there is only one basic block in the program.
        Multiple basic blocks may have non-conditional control flow among them,
        in which the execution order is deterministic and does not depend on program state.
        This may be a sequence of basic blocks with fixed `JUMP`s or without explicit terminators.
        """

class FrameSet:
    r"""
    A collection of Quil frames (`DEFFRAME` instructions) with utility methods.
    """
    def __eq__(self, other:builtins.object) -> builtins.bool: ...
    def __len__(self) -> builtins.int:
        r"""
        Return the number of frames described within.
        """
    def __new__(cls) -> FrameSet: ...
    def __repr__(self) -> builtins.str: ...
    def get(self, identifier:FrameIdentifier) -> typing.Optional[builtins.dict[builtins.str, AttributeValue]]:
        r"""
        Retrieve the attributes of a frame by its identifier.
        """
    def get_all_frames(self) -> builtins.dict[FrameIdentifier, builtins.dict[builtins.str, AttributeValue]]: ...
    def get_keys(self) -> builtins.list[FrameIdentifier]:
        r"""
        Return a list of all ``FrameIdentifier``s described by this ``FrameSet``.
        """
    def insert(self, identifier:FrameIdentifier, attributes:typing.Mapping[builtins.str, AttributeValue]) -> None:
        r"""
        Insert a new frame by ID, overwriting any existing one.
        """
    def intersection(self, identifiers:builtins.set[FrameIdentifier]) -> FrameSet:
        r"""
        Return a new [`FrameSet`] which describes only the given [`FrameIdentifier`]s.
        """
    def is_empty(self) -> builtins.bool:
        r"""
        Return true if this describes no frames.
        """
    def merge(self, other:FrameSet) -> None:
        r"""
        Merge another [FrameSet] with this one, overwriting any existing keys
        """
    def to_instructions(self) -> builtins.list[Instruction]:
        r"""
        Return the Quil instructions which describe the contained frames.
        """

class MaybeCalibrationExpansion:
    r"""
    The result of an attempt to expand an instruction within a [`Program`]
    """
    def __getnewargs__(self) -> tuple[CalibrationExpansion, InstructionIndex]: ...
    def __repr__(self) -> builtins.str: ...
    class Expanded(MaybeCalibrationExpansion):
        r"""
        The instruction was expanded into others
        """
        __match_args__ = ("_0",)
        @property
        def _0(self) -> CalibrationExpansion: ...
        def __getitem__(self, key:builtins.int) -> typing.Any: ...
        def __len__(self) -> builtins.int: ...
        def __new__(cls, _0:CalibrationExpansion) -> MaybeCalibrationExpansion.Expanded: ...
    
    class Unexpanded(MaybeCalibrationExpansion):
        r"""
        The instruction was not expanded, but was simply copied over into the target program at the given instruction index
        """
        __match_args__ = ("_0",)
        @property
        def _0(self) -> builtins.int: ...
        def __getitem__(self, key:builtins.int) -> typing.Any: ...
        def __len__(self) -> builtins.int: ...
        def __new__(cls, _0:builtins.int) -> MaybeCalibrationExpansion.Unexpanded: ...
    

class MemoryRegion:
    @property
    def sharing(self) -> typing.Optional[Sharing]: ...
    @property
    def size(self) -> Vector: ...
    def __eq__(self, other:builtins.object) -> builtins.bool: ...
    def __getnewargs__(self) -> tuple[Vector, typing.Optional[Sharing]]: ...
    def __hash__(self) -> builtins.int: ...
    def __new__(cls, size:Vector, sharing:typing.Optional[Sharing]) -> MemoryRegion: ...
    def __repr__(self) -> builtins.str: ...

class Program:
    r"""
    A Quil Program instance describes a quantum program with metadata used in execution.
    
    This contains not only instructions which are executed in turn on the quantum processor, but
    also the "headers" used to describe and manipulate those instructions, such as calibrations
    and frame definitions.
    """
    @property
    def body_instructions(self) -> builtins.list[Instruction]: ...
    @property
    def calibrations(self) -> CalibrationSet: ...
    @property
    def declarations(self) -> builtins.dict[builtins.str, Declaration]: ...
    @property
    def frames(self) -> FrameSet: ...
    @property
    def gate_definitions(self) -> builtins.dict[builtins.str, GateDefinition]: ...
    @property
    def instructions(self) -> builtins.list[Instruction]: ...
    @property
    def memory_regions(self) -> builtins.dict[builtins.str, MemoryRegion]: ...
    @property
    def pragma_extern_map(self) -> builtins.set[typing.Optional[builtins.str], Pragma]: ...
    @property
    def used_qubits(self) -> builtins.set[Qubit]: ...
    @property
    def waveforms(self) -> builtins.dict[builtins.str, Waveform]: ...
    @calibrations.setter
    def calibrations(self, value: CalibrationSet) -> None: ...
    @frames.setter
    def frames(self, value: FrameSet) -> None: ...
    @gate_definitions.setter
    def gate_definitions(self, value: builtins.dict[builtins.str, GateDefinition]) -> None: ...
    @instructions.setter
    def instructions(self, value: builtins.list[Instruction]) -> None: ...
    @memory_regions.setter
    def memory_regions(self, value: builtins.dict[builtins.str, MemoryRegion]) -> None: ...
    @waveforms.setter
    def waveforms(self, value: builtins.dict[builtins.str, Waveform]) -> None: ...
    def __add__(self, rhs:Program) -> Program: ...
    def __eq__(self, other:builtins.object) -> builtins.bool: ...
    def __getstate__(self) -> bytes:
        r"""
        This will raise an error if the program contains any unresolved
        placeholders. This is because they can't be converted to valid quil,
        nor can they be serialized and deserialized in a consistent way.
        """
    def __iadd__(self, rhs:Program) -> None: ...
    def __new__(cls) -> Program: ...
    def __repr__(self) -> builtins.str: ...
    def __setstate__(self, state:bytes) -> None: ...
    def add_instruction(self, instruction:Instruction) -> None:
        r"""
        Add an instruction to the end of the program.
        
        Note, parsing extern signatures is deferred here to maintain infallibility
        of [`Program::add_instruction`]. This means that invalid `PRAGMA EXTERN`
        instructions are still added to the [`Program::extern_pragma_map`];
        duplicate `PRAGMA EXTERN` names are overwritten.
        """
    def add_instructions(self, instructions:typing.Sequence[Instruction]) -> None:
        r"""
        Add a list of instructions to the end of the program.
        """
    def clone_without_body_instructions(self) -> Program:
        r"""
        Return a deep copy of the `Program`, but without the body instructions.
        """
    def control_flow_graph(self) -> ControlFlowGraph:
        r"""
        Return the [control flow graph][] of the program.
        
        [control flow graph]: https://en.wikipedia.org/wiki/Control-flow_graph
        """
    def copy(self) -> Program:
        r"""
        Return a deep copy of the `Program`.
        """
    def dagger(self) -> Program:
        r"""
        Creates a new conjugate transpose of the [`Program`] by reversing the order of gate
        instructions and applying the DAGGER modifier to each.
        
        # Errors
        
        Errors if any of the instructions in the program are not [`Instruction::Gate`]
        """
    def expand_calibrations(self) -> Program:
        r"""
        Expand any instructions in the program which have a matching calibration,
        leaving the others unchanged.
        Return the expanded copy of the program.
        
        Returns an error if any instruction expands into itself.
        
        See [`Program::expand_calibrations_with_source_map`] for a version that returns a source mapping.
        """
    def expand_calibrations_with_source_map(self) -> ProgramCalibrationExpansion:
        r"""
        Expand any instructions in the program which have a matching calibration,
        leaving the others unchanged.
        Return the expanded copy of the program
        and a source mapping describing the expansions made.
        """
    def filter_instructions(self, predicate:collections.abc.Callable[[Instruction], bool]) -> Program:
        r"""
        Return a new ``Program`` containing only the instructions
        for which `predicate` returns true.
        """
    def into_simplified(self) -> Program:
        r"""
        Simplify this program into a new [`Program`] which contains only instructions
        and definitions which are executed; effectively, perform dead code removal.
        
        Removes:
        - All calibrations, following calibration expansion
        - Frame definitions which are not used by any instruction such as `PULSE` or `CAPTURE`
        - Waveform definitions which are not used by any instruction
        - `PRAGMA EXTERN` instructions which are not used by any `CALL` instruction (see
          [`Program::extern_pragma_map`]).
        
        When a valid program is simplified, it remains valid.
        
        # Note
        
        If you need custom instruction handling during simplification,
        use [`InstructionHandler::simplify_program`] instead.
        """
    @staticmethod
    def parse(input:builtins.str) -> Program:
        r"""
        Parse a ``Program`` from a string.
        
        Raises a ``ProgramError`` if the string isn't a valid Quil expression.
        """
    def resolve_placeholders(self) -> None:
        r"""
        Resolve [`LabelPlaceholder`]s and [`QubitPlaceholder`]s within the program using default resolvers.
        
        See [`resolve_placeholders_with_custom_resolvers`](Self::resolve_placeholders_with_custom_resolvers),
        [`default_target_resolver`](Self::default_target_resolver),
        and [`default_qubit_resolver`](Self::default_qubit_resolver) for more information.
        """
    def resolve_placeholders_with_custom_resolvers(self, *, target_resolver:typing.Optional[collections.abc.Callable[[TargetPlaceholder], str | None]]=None, qubit_resolver:typing.Optional[collections.abc.Callable[[QubitPlaceholder], int | None]]=None) -> None:
        r"""
        Resolve ``TargetPlaceholder``s and ``QubitPlaceholder``s within the program.
        
        The resolved values will remain unique to that placeholder
        within the scope of the program.
        If you provide ``target_resolver`` and/or ``qubit_resolver``,
        those will be used to resolve those values respectively.
        If your resolver returns `None` for a particular placeholder,
        it will not be replaced but will be left as a placeholder.
        If you do not provide a resolver for a placeholder,
        a default resolver will be used which will generate
        a unique value for that placeholder within the scope of the program
        using an auto-incrementing value (for qubit) or suffix (for target)
        while ensuring that unique value is not already in use within the program.
        """
    def to_instructions(self) -> builtins.list[Instruction]:
        r"""
        Return a copy of all of the instructions which constitute this [`Program`].
        """
    def to_quil(self) -> builtins.str: ...
    def to_quil_or_debug(self) -> builtins.str: ...
    def to_unitary(self, n_qubits:builtins.int) -> numpy.typing.NDArray[numpy.complex128]:
        r"""
        Return the unitary of a program.
        
        # Errors
        
        Returns an error if the program contains instructions other than `Gate`s.
        """
    def wrap_in_loop(self, loop_count_reference:MemoryReference, start_target:Target, end_target:Target, iterations:builtins.int) -> Program:
        r"""
        Return a copy of the [`Program`] wrapped in a loop that repeats `iterations` times.
        
        The loop is constructed by wrapping the body of the program in classical Quil instructions.
        The given `loop_count_reference` must refer to an INTEGER memory region. The value at the
        reference given will be set to `iterations` and decremented in the loop. The loop will
        terminate when the reference reaches 0. For this reason your program should not itself
        modify the value at the reference unless you intend to modify the remaining number of
        iterations (i.e. to break the loop).
        
        The given `start_target` and `end_target` will be used as the entry and exit points for the
        loop, respectively. You should provide unique [`Target`]s that won't be used elsewhere in
        the program.
        
        If `iterations` is 0, then a copy of the program is returned without any changes.
        """

class ProgramCalibrationExpansion:
    @property
    def program(self) -> Program:
        r"""
        The program containing the instructions.
        """
    @property
    def source_map(self) -> ProgramCalibrationExpansionSourceMap:
        r"""
        The source mapping describing the expansions made.
        """
    def __eq__(self, other:builtins.object) -> builtins.bool: ...
    def __repr__(self) -> builtins.str: ...

class ProgramCalibrationExpansionSourceMap:
    def __repr__(self) -> builtins.str: ...
    def entries(self) -> builtins.list[ProgramCalibrationExpansionSourceMapEntry]: ...
    def list_sources_for_calibration_used(self, calibration_used:CalibrationSource) -> builtins.list[builtins.int]:
        r"""
        Given a particular calibration (`DEFCAL` or `DEFCAL MEASURE`), =
        return the locations in the source which were expanded using that calibration.
        
        This is `O(n)` where `n` is the number of first-level calibration expansions performed.
        """
    def list_sources_for_target_index(self, target_index:builtins.int) -> builtins.list[builtins.int]:
        r"""
        Return all source ranges in the source map
        which were used to generate the target index.
        
        This is `O(n)` where `n` is the number of entries in the map.
        """
    def list_targets_for_source_index(self, source_index:builtins.int) -> builtins.list[MaybeCalibrationExpansion]:
        r"""
        Return all target ranges which were used to generate the source range.
        
        This is `O(n)` where `n` is the number of entries in the map.
        """

class ProgramCalibrationExpansionSourceMapEntry:
    def __repr__(self) -> builtins.str: ...
    def source_location(self) -> builtins.int:
        r"""
        The instruction index within the source program's body instructions.
        """
    def target_location(self) -> MaybeCalibrationExpansion:
        r"""
        The location of the expanded instruction within the target
        program's body instructions.
        """

class ProgramError(builtins.QuilError):
    r"""
    Errors encountered related to a Program.
    """
    ...

class QubitGraphError(builtins.ProgramError):
    ...

class ScheduleSeconds:
    r"""
    A Schedule is a ``DependencyGraph`` flattened into a linear sequence of instructions,
    each of which is assigned a start time and duration.
    """
    @property
    def duration(self) -> builtins.float:
        r"""
        The schedule duration, in seconds.
        
        This is the maximum end time among all scheduled items.
        """
    @property
    def items(self) -> builtins.list[ScheduleSecondsItem]:
        r"""
        Scheduled items, in an unspecified order.
        """
    def __eq__(self, other:builtins.object) -> builtins.bool: ...
    def __repr__(self) -> builtins.str: ...

class ScheduleSecondsItem:
    r"""
    A single item within a schedule, representing a single instruction within a basic block.
    """
    @property
    def instruction_index(self) -> builtins.int:
        r"""
        The time span during which the instruction is scheduled.
        """
    @property
    def time_span(self) -> TimeSpanSeconds:
        r"""
        The index of the instruction within the basic block.
        """
    def __eq__(self, other:builtins.object) -> builtins.bool: ...
    def __repr__(self) -> builtins.str: ...

class TimeSpanSeconds:
    r"""
    A time span, in seconds.
    """
    @property
    def duration(self) -> builtins.float:
        r"""
        The duration of the time span, in seconds.
        """
    @property
    def end(self) -> builtins.float:
        r"""
        The end time of the time span, in seconds.
        
        This is the sum of the start time and duration.
        """
    @property
    def start(self) -> builtins.float:
        r"""
        The inclusive start time of the time span,
        in seconds relative to the start of the scheduling context (such as the basic block).
        """
    def __eq__(self, other:builtins.object) -> builtins.bool: ...
    def __repr__(self) -> builtins.str: ...




# This file is automatically generated by pyo3_stub_gen
# ruff: noqa: E501, F401

import builtins
import quil.instructions
import typing
from enum import Enum

class EvaluationError(builtins.QuilError):
    r"""
    Errors that may occur while evaluation an ``Expression``.
    """
    ...

class Expression:
    r"""
    The type of Quil expressions.
    
    Quil expressions take advantage of *structural sharing*; if a Quil expression contains the same
    subexpression twice, such as `x + y` in `(x + y) * (x + y)`, the two children of the `*` node
    will be the *same pointer*.  This is implemented through *interning*, also known as
    *hash-consing*; the recursive references to child nodes are done via [`ArcIntern<Expression>`]s.
    Creating an [`ArcIntern`] from an `Expression` will always return the same pointer for the same
    expression.
    
    The structural sharing means that equality, cloning, and hashing on Quil expressions are all
    very cheap, as they do not need to be recursive: equality of [`ArcIntern`]s is a single-pointer
    comparison, cloning of [`ArcIntern`]s is a pointer copy and an atomic increment, and hashing of
    [`ArcIntern`]s hashes a single pointer.  It is also very cheap to key [`HashMap`]s by an
    [`ArcIntern<Expression>`], which can allow for cheap memoization of operations on `Expression`s.
    
    The structural sharing also means that Quil expressions are fundamentally *immutable*; it is
    *impossible* to get an owned or `&mut` reference to the child `Expression`s of any `Expression`,
    as the use of interning means there may be multiple references to that `Expression` at any time.
    
    Note that when comparing Quil expressions, any embedded NaNs are treated as *equal* to other
    NaNs, not unequal, in contravention of the IEEE 754 spec.
    """
    def __add__(self, other:Expression) -> Expression: ...
    def __getnewargs__(self) -> tuple[MemoryReference | FunctionCallExpression | InfixExpression | complex | PrefixExpression | str]: ...
    def __mul__(self, other:Expression) -> Expression: ...
    def __repr__(self) -> builtins.str: ...
    def __sub__(self, other:Expression) -> Expression: ...
    def __truediv__(self, other:Expression) -> Expression: ...
    def evaluate(self, variables:typing.Mapping[builtins.str, builtins.complex], memory_references:typing.Mapping[builtins.str, typing.Sequence[builtins.float]]) -> builtins.complex:
        r"""
        Evaluate an expression, expecting that it may be fully reduced to a single complex number.
        
        If it cannot be reduced to a complex number, this raises an error.
        """
    def into_simplified(self) -> Expression:
        r"""
        Return an expression derived from this one, simplified as much as possible.
        """
    @staticmethod
    def parse(input:builtins.str) -> Expression:
        r"""
        Parse an ``Expression`` from a string.
        
        Raises a ``ParseExpressionError`` error if the string isn't a valid Quil expression.
        """
    def substitute_variables(self, variable_values:typing.Mapping[builtins.str, Expression]) -> Expression:
        r"""
        Substitute an expression in the place of each matching variable.
        """
    def to_real(self) -> builtins.float:
        r"""
        If this is a number with imaginary part "equal to" zero (of _small_ absolute value), return
        that number. Otherwise, error with an evaluation error of a descriptive type.
        """
    class Address(Expression):
        __match_args__ = ("_0",)
        @property
        def _0(self) -> MemoryReference: ...
        def __getitem__(self, key:builtins.int) -> typing.Any: ...
        def __len__(self) -> builtins.int: ...
        def __new__(cls, _0:MemoryReference) -> Expression.Address: ...
    
    class FunctionCall(Expression):
        __match_args__ = ("_0",)
        @property
        def _0(self) -> FunctionCallExpression: ...
        def __getitem__(self, key:builtins.int) -> typing.Any: ...
        def __len__(self) -> builtins.int: ...
        def __new__(cls, _0:FunctionCallExpression) -> Expression.FunctionCall: ...
    
    class Infix(Expression):
        __match_args__ = ("_0",)
        @property
        def _0(self) -> InfixExpression: ...
        def __getitem__(self, key:builtins.int) -> typing.Any: ...
        def __len__(self) -> builtins.int: ...
        def __new__(cls, _0:InfixExpression) -> Expression.Infix: ...
    
    class Number(Expression):
        __match_args__ = ("_0",)
        @property
        def _0(self) -> builtins.complex: ...
        def __getitem__(self, key:builtins.int) -> typing.Any: ...
        def __len__(self) -> builtins.int: ...
        def __new__(cls, _0:builtins.complex) -> Expression.Number: ...
    
    class Pi(Expression):
        __match_args__ = ((),)
        def __getitem__(self, key:builtins.int) -> typing.Any: ...
        def __len__(self) -> builtins.int: ...
        def __new__(cls) -> Expression.Pi: ...
    
    class Prefix(Expression):
        __match_args__ = ("_0",)
        @property
        def _0(self) -> PrefixExpression: ...
        def __getitem__(self, key:builtins.int) -> typing.Any: ...
        def __len__(self) -> builtins.int: ...
        def __new__(cls, _0:PrefixExpression) -> Expression.Prefix: ...
    
    class Variable(Expression):
        __match_args__ = ("_0",)
        @property
        def _0(self) -> builtins.str: ...
        def __getitem__(self, key:builtins.int) -> typing.Any: ...
        def __len__(self) -> builtins.int: ...
        def __new__(cls, _0:builtins.str) -> Expression.Variable: ...
    

class FunctionCallExpression:
    r"""
    The type of function call Quil expressions, e.g. `sin(e)`.
    
    Quil expressions take advantage of *structural sharing*, which is why the `expression` here is
    wrapped in an [`ArcIntern`]; for more details, see the documentation for [`Expression`].
    
    Note that when comparing Quil expressions, any embedded NaNs are treated as *equal* to other
    NaNs, not unequal, in contravention of the IEEE 754 spec.
    """
    @property
    def expression(self) -> Expression: ...
    @property
    def function(self) -> ExpressionFunction: ...
    def __eq__(self, other:builtins.object) -> builtins.bool: ...
    def __getnewargs__(self) -> tuple[ExpressionFunction, Expression]: ...
    def __hash__(self) -> builtins.int: ...
    def __new__(cls, function:ExpressionFunction, expression:Expression) -> FunctionCallExpression: ...
    def __repr__(self) -> builtins.str: ...

class InfixExpression:
    r"""
    The type of infix Quil expressions, e.g. `e1 + e2`.
    
    Quil expressions take advantage of *structural sharing*, which is why the `left` and `right`
    expressions here are wrapped in [`ArcIntern`]s; for more details, see the documentation for
    [`Expression`].
    
    Note that when comparing Quil expressions, any embedded NaNs are treated as *equal* to other
    NaNs, not unequal, in contravention of the IEEE 754 spec.
    """
    @property
    def left(self) -> Expression: ...
    @property
    def operator(self) -> InfixOperator: ...
    @property
    def right(self) -> Expression: ...
    def __eq__(self, other:builtins.object) -> builtins.bool: ...
    def __getnewargs__(self) -> tuple[Expression, InfixOperator, Expression]: ...
    def __hash__(self) -> builtins.int: ...
    def __new__(cls, left:Expression, operator:InfixOperator, right:Expression) -> InfixExpression: ...
    def __repr__(self) -> builtins.str: ...

class ParseExpressionError(builtins.QuilError):
    r"""
    Errors that may occur while parsing an ``Expression``.
    """
    ...

class PrefixExpression:
    r"""
    The type of prefix Quil expressions, e.g. `-e`.
    
    Quil expressions take advantage of *structural sharing*, which is why the `expression` here is
    wrapped in an [`ArcIntern`]; for more details, see the documentation for [`Expression`].
    
    Note that when comparing Quil expressions, any embedded NaNs are treated as *equal* to other
    NaNs, not unequal, in contravention of the IEEE 754 spec.
    """
    @property
    def expression(self) -> Expression: ...
    @property
    def operator(self) -> PrefixOperator: ...
    def __eq__(self, other:builtins.object) -> builtins.bool: ...
    def __getnewargs__(self) -> tuple[PrefixOperator, Expression]: ...
    def __hash__(self) -> builtins.int: ...
    def __new__(cls, operator:PrefixOperator, expression:Expression) -> PrefixExpression: ...
    def __repr__(self) -> builtins.str: ...

class ExpressionFunction(Enum):
    r"""
    A function defined within Quil syntax.
    """
    Cis = ...
    Cosine = ...
    Exponent = ...
    Sine = ...
    SquareRoot = ...

    def __getnewargs__(self) -> tuple[builtins.int]: ...
    def __new__(cls, value:builtins.int) -> ExpressionFunction: ...
    def __repr__(self) -> builtins.str: ...

class InfixOperator(Enum):
    Caret = ...
    Plus = ...
    Minus = ...
    Slash = ...
    Star = ...

    def __getnewargs__(self) -> tuple[builtins.int]: ...
    def __new__(cls, value:builtins.int) -> InfixOperator: ...
    def __repr__(self) -> builtins.str: ...

class PrefixOperator(Enum):
    Plus = ...
    Minus = ...

    def __getnewargs__(self) -> tuple[builtins.int]: ...
    def __new__(cls, value:builtins.int) -> PrefixOperator: ...
    def __repr__(self) -> builtins.str: ...




from quil import waveforms


class TestConstructors:
    """Test that class constructors exist,
    accept the expected argument names,
    and correctly round-trip their values.
    """

    def test_boxcar_kernel(self, phase: float, scale: float, sample_count: int):
        wf = waveforms.BoxcarKernel(phase=phase, scale=scale, sample_count=sample_count)
        assert wf is not None
        assert wf.phase == phase
        assert wf.scale == scale
        assert wf.sample_count == sample_count

    def test_erf_square(
        self,
        duration: float,
        risetime: float,
        sample_rate: float,
        pad_left: float,
        pad_right: float,
        positive_polarity: bool,
        scale: float,
        phase: float,
        detuning: float,
    ):
        wf = waveforms.ErfSquare(
            duration=duration,
            risetime=risetime,
            sample_rate=sample_rate,
            pad_left=pad_left,
            pad_right=pad_right,
            positive_polarity=positive_polarity,
            scale=scale,
            phase=phase,
            detuning=detuning,
        )
        assert wf is not None
        assert wf.duration == duration
        assert wf.risetime == risetime
        assert wf.sample_rate == sample_rate
        assert wf.pad_left == pad_left
        assert wf.pad_right == pad_right
        assert wf.positive_polarity == positive_polarity
        assert wf.scale == scale
        assert wf.phase == phase
        assert wf.detuning == detuning

    def test_gaussian(
        self,
        duration: float,
        fwhm: float,
        t0: float,
        sample_rate: float,
        scale: float,
        phase: float,
        detuning: float,
    ):
        wf = waveforms.Gaussian(
            duration=duration,
            fwhm=fwhm,
            t0=t0,
            sample_rate=sample_rate,
            scale=scale,
            phase=phase,
            detuning=detuning,
        )
        assert wf is not None
        assert wf.duration == duration
        assert wf.fwhm == fwhm
        assert wf.t0 == t0
        assert wf.sample_rate == sample_rate
        assert wf.scale == scale
        assert wf.phase == phase
        assert wf.detuning == detuning

    def test_drag_gaussian(
        self,
        duration: float,
        fwhm: float,
        t0: float,
        anh: float,
        alpha: float,
        sample_rate: float,
        scale: float,
        phase: float,
        detuning: float,
    ):
        wf = waveforms.DragGaussian(
            duration=duration,
            fwhm=fwhm,
            t0=t0,
            anh=anh,
            alpha=alpha,
            sample_rate=sample_rate,
            scale=scale,
            phase=phase,
            detuning=detuning,
        )
        assert wf is not None
        assert wf.duration == duration
        assert wf.fwhm == fwhm
        assert wf.t0 == t0
        assert wf.anh == anh
        assert wf.alpha == alpha
        assert wf.sample_rate == sample_rate
        assert wf.scale == scale
        assert wf.phase == phase
        assert wf.detuning == detuning

    def test_hermite_gaussian(
        self,
        duration: float,
        fwhm: float,
        t0: float,
        anh: float,
        alpha: float,
        sample_rate: float,
        second_order_hrm_coeff: float,
        scale: float,
        phase: float,
        detuning: float,
    ):
        wf = waveforms.HermiteGaussian(
            duration=duration,
            fwhm=fwhm,
            t0=t0,
            anh=anh,
            alpha=alpha,
            sample_rate=sample_rate,
            second_order_hrm_coeff=second_order_hrm_coeff,
            scale=scale,
            phase=phase,
            detuning=detuning,
        )
        assert wf is not None
        assert wf.duration == duration
        assert wf.fwhm == fwhm
        assert wf.t0 == t0
        assert wf.anh == anh
        assert wf.alpha == alpha
        assert wf.sample_rate == sample_rate
        assert wf.second_order_hrm_coeff == second_order_hrm_coeff
        assert wf.scale == scale
        assert wf.phase == phase
        assert wf.detuning == detuning



# This file is automatically generated by pyo3_stub_gen
# ruff: noqa: E501, F401

import builtins
import numpy
import numpy.typing
import quil.expression
import typing
from enum import Enum

class Arithmetic:
    @property
    def destination(self) -> MemoryReference: ...
    @property
    def operator(self) -> ArithmeticOperator: ...
    @property
    def source(self) -> ArithmeticOperand: ...
    def __eq__(self, other:builtins.object) -> builtins.bool: ...
    def __getnewargs__(self) -> tuple[ArithmeticOperator, MemoryReference, ArithmeticOperand]: ...
    def __hash__(self) -> builtins.int: ...
    def __new__(cls, operator:ArithmeticOperator, destination:MemoryReference, source:ArithmeticOperand) -> Arithmetic: ...
    def __repr__(self) -> builtins.str: ...
    def to_quil(self) -> builtins.str: ...
    def to_quil_or_debug(self) -> builtins.str: ...

class ArithmeticOperand:
    def __getnewargs__(self) -> tuple[int | float | MemoryReference]: ...
    def __repr__(self) -> builtins.str: ...
    def to_quil(self) -> builtins.str: ...
    def to_quil_or_debug(self) -> builtins.str: ...
    class LiteralInteger(ArithmeticOperand):
        __match_args__ = ("_0",)
        @property
        def _0(self) -> builtins.int: ...
        def __getitem__(self, key:builtins.int) -> typing.Any: ...
        def __len__(self) -> builtins.int: ...
        def __new__(cls, _0:builtins.int) -> ArithmeticOperand.LiteralInteger: ...
    
    class LiteralReal(ArithmeticOperand):
        __match_args__ = ("_0",)
        @property
        def _0(self) -> builtins.float: ...
        def __getitem__(self, key:builtins.int) -> typing.Any: ...
        def __len__(self) -> builtins.int: ...
        def __new__(cls, _0:builtins.float) -> ArithmeticOperand.LiteralReal: ...
    
    class MemoryReference(ArithmeticOperand):
        __match_args__ = ("_0",)
        @property
        def _0(self) -> MemoryReference: ...
        def __getitem__(self, key:builtins.int) -> typing.Any: ...
        def __len__(self) -> builtins.int: ...
        def __new__(cls, _0:MemoryReference) -> ArithmeticOperand.MemoryReference: ...
    

class AttributeValue:
    def __getnewargs__(self) -> tuple[str | Expression]: ...
    def __repr__(self) -> builtins.str: ...
    def to_quil(self) -> builtins.str: ...
    def to_quil_or_debug(self) -> builtins.str: ...
    class Expression(AttributeValue):
        __match_args__ = ("_0",)
        @property
        def _0(self) -> Expression: ...
        def __getitem__(self, key:builtins.int) -> typing.Any: ...
        def __len__(self) -> builtins.int: ...
        def __new__(cls, _0:Expression) -> AttributeValue.Expression: ...
    
    class String(AttributeValue):
        __match_args__ = ("_0",)
        @property
        def _0(self) -> builtins.str: ...
        def __getitem__(self, key:builtins.int) -> typing.Any: ...
        def __len__(self) -> builtins.int: ...
        def __new__(cls, _0:builtins.str) -> AttributeValue.String: ...
    

class BinaryLogic:
    @property
    def destination(self) -> MemoryReference: ...
    @property
    def operator(self) -> BinaryOperator: ...
    @property
    def source(self) -> BinaryOperand: ...
    def __eq__(self, other:builtins.object) -> builtins.bool: ...
    def __getnewargs__(self) -> tuple[BinaryOperator, MemoryReference, BinaryOperand]: ...
    def __hash__(self) -> builtins.int: ...
    def __new__(cls, operator:BinaryOperator, destination:MemoryReference, source:BinaryOperand) -> BinaryLogic: ...
    def __repr__(self) -> builtins.str: ...
    def to_quil(self) -> builtins.str: ...
    def to_quil_or_debug(self) -> builtins.str: ...

class CalibrationDefinition:
    @property
    def identifier(self) -> CalibrationIdentifier: ...
    @property
    def instructions(self) -> builtins.list[Instruction]: ...
    @property
    def modifiers(self) -> builtins.list[GateModifier]:
        r"""
        The list of [`GateModifier`]s that this calibration definition is for.
        """
    @property
    def name(self) -> builtins.str:
        r"""
        The gate name that this calibration definition is for.
        """
    @property
    def parameters(self) -> builtins.list[Expression]:
        r"""
        The list of parameters that this calibration definition is for.
        """
    @property
    def qubits(self) -> builtins.list[Qubit]:
        r"""
        The list of [`Qubit`]s that this calibration definition is for.
        """
    @identifier.setter
    def identifier(self, value: CalibrationIdentifier) -> None: ...
    @instructions.setter
    def instructions(self, value: builtins.list[Instruction]) -> None: ...
    def __eq__(self, other:builtins.object) -> builtins.bool: ...
    def __getnewargs__(self) -> tuple[CalibrationIdentifier, builtins.list[Instruction]]: ...
    def __new__(cls, identifier:CalibrationIdentifier, instructions:typing.Sequence[Instruction]) -> CalibrationDefinition:
        r"""
        Builds a new calibration definition.
        """
    def __repr__(self) -> builtins.str: ...
    def to_quil(self) -> builtins.str: ...
    def to_quil_or_debug(self) -> builtins.str: ...

class CalibrationIdentifier:
    r"""
    Unique identifier for a calibration definition within a program
    """
    @property
    def modifiers(self) -> builtins.list[GateModifier]:
        r"""
        The modifiers applied to the gate
        """
    @property
    def name(self) -> builtins.str:
        r"""
        The name of the gate
        """
    @property
    def parameters(self) -> builtins.list[Expression]:
        r"""
        The parameters of the gate - these are the variables in the calibration definition
        """
    @property
    def qubits(self) -> builtins.list[Qubit]:
        r"""
        The qubits on which the gate is applied
        """
    @modifiers.setter
    def modifiers(self, value: builtins.list[GateModifier]) -> None:
        r"""
        The modifiers applied to the gate
        """
    @name.setter
    def name(self, value: builtins.str) -> None:
        r"""
        The name of the gate
        """
    @parameters.setter
    def parameters(self, value: builtins.list[Expression]) -> None:
        r"""
        The parameters of the gate - these are the variables in the calibration definition
        """
    @qubits.setter
    def qubits(self, value: builtins.list[Qubit]) -> None:
        r"""
        The qubits on which the gate is applied
        """
    def __eq__(self, other:builtins.object) -> builtins.bool: ...
    def __getnewargs__(self) -> tuple[builtins.str, builtins.list[Expression], builtins.list[Qubit], builtins.list[GateModifier]]: ...
    def __new__(cls, name:builtins.str, parameters:typing.Sequence[Expression], qubits:typing.Sequence[Qubit], modifiers:typing.Sequence[GateModifier]) -> CalibrationIdentifier:
        r"""
        Builds a new calibration identifier.
        
        Raises an error if the given name isn't a valid Quil identifier.
        """
    def __repr__(self) -> builtins.str: ...
    def to_quil(self) -> builtins.str: ...
    def to_quil_or_debug(self) -> builtins.str: ...

class Call:
    r"""
    A call instruction with a name and arguments.
    
    An instruction that calls an external function declared with a `PRAGMA EXTERN` instruction.
    These calls are generally specific to a particular hardware or virtual machine backend.
    
    For further detail, see:
    
    * [Other instructions and Directives](https://github.com/quil-lang/quil/blob/master/rfcs/extern-call.md) in the Quil specification.
    * [EXTERN / CALL RFC](https://github.com/quil-lang/quil/blob/master/rfcs/extern-call.md)
    * [quil#87](https://github.com/quil-lang/quil/issues/87)
    
    Also see [`ExternSignature`].
    """
    @property
    def arguments(self) -> builtins.list[CallArgument]:
        r"""
        The arguments of the call instruction.
        """
    @property
    def name(self) -> builtins.str:
        r"""
        The name of the call instruction. This must be a valid user identifier.
        """
    def __eq__(self, other:builtins.object) -> builtins.bool: ...
    def __getnewargs__(self) -> tuple[builtins.str, builtins.list[CallArgument]]: ...
    def __hash__(self) -> builtins.int: ...
    def __new__(cls, name:builtins.str, arguments:typing.Sequence[CallArgument]) -> Call:
        r"""
        Create a new call instruction with resolved arguments.
        This will validate the name as a user identifier.
        """
    def __repr__(self) -> builtins.str: ...
    def to_quil(self) -> builtins.str: ...
    def to_quil_or_debug(self) -> builtins.str: ...

class CallArgument:
    r"""
    A parsed, but unresolved call argument. This may be resolved into a [`ResolvedCallArgument`]
    with the appropriate [`ExternSignature`]. Resolution is required for building the
    [`crate::Program`] memory graph.
    """
    def __getnewargs__(self) -> tuple[str | MemoryReference | complex]: ...
    def __repr__(self) -> builtins.str: ...
    def to_quil(self) -> builtins.str: ...
    def to_quil_or_debug(self) -> builtins.str: ...
    class Identifier(CallArgument):
        r"""
        A reference to a declared memory location. Note, this may be resolved to either
        a scalar or vector. In the former case, the assumed index is 0.
        """
        __match_args__ = ("_0",)
        @property
        def _0(self) -> builtins.str: ...
        def __getitem__(self, key:builtins.int) -> typing.Any: ...
        def __len__(self) -> builtins.int: ...
        def __new__(cls, _0:builtins.str) -> CallArgument.Identifier: ...
    
    class Immediate(CallArgument):
        r"""
        An immediate value. This may be resolved to a non-mutable scalar.
        """
        __match_args__ = ("_0",)
        @property
        def _0(self) -> builtins.complex: ...
        def __getitem__(self, key:builtins.int) -> typing.Any: ...
        def __len__(self) -> builtins.int: ...
        def __new__(cls, _0:builtins.complex) -> CallArgument.Immediate: ...
    
    class MemoryReference(CallArgument):
        r"""
        A reference to a memory location. This may be resolved to a scalar.
        """
        __match_args__ = ("_0",)
        @property
        def _0(self) -> MemoryReference: ...
        def __getitem__(self, key:builtins.int) -> typing.Any: ...
        def __len__(self) -> builtins.int: ...
        def __new__(cls, _0:MemoryReference) -> CallArgument.MemoryReference: ...
    

class CallError(builtins.QuilError):
    r"""
    Errors that may occur when initializing a ``Call``.
    """
    ...

class Capture:
    @property
    def blocking(self) -> builtins.bool: ...
    @property
    def frame(self) -> FrameIdentifier: ...
    @property
    def memory_reference(self) -> MemoryReference: ...
    @property
    def waveform(self) -> WaveformInvocation: ...
    @blocking.setter
    def blocking(self, value: builtins.bool) -> None: ...
    @frame.setter
    def frame(self, value: FrameIdentifier) -> None: ...
    @memory_reference.setter
    def memory_reference(self, value: MemoryReference) -> None: ...
    @waveform.setter
    def waveform(self, value: WaveformInvocation) -> None: ...
    def __eq__(self, other:builtins.object) -> builtins.bool: ...
    def __getnewargs__(self) -> tuple[builtins.bool, FrameIdentifier, MemoryReference, WaveformInvocation]: ...
    def __new__(cls, blocking:builtins.bool, frame:FrameIdentifier, memory_reference:MemoryReference, waveform:WaveformInvocation) -> Capture: ...
    def __repr__(self) -> builtins.str: ...
    def to_quil(self) -> builtins.str: ...
    def to_quil_or_debug(self) -> builtins.str: ...

class CircuitDefinition:
    @property
    def instructions(self) -> builtins.list[Instruction]: ...
    @property
    def name(self) -> builtins.str: ...
    @property
    def parameters(self) -> builtins.list[builtins.str]: ...
    @property
    def qubit_variables(self) -> builtins.list[builtins.str]: ...
    @instructions.setter
    def instructions(self, value: builtins.list[Instruction]) -> None: ...
    @name.setter
    def name(self, value: builtins.str) -> None: ...
    @parameters.setter
    def parameters(self, value: builtins.list[builtins.str]) -> None: ...
    @qubit_variables.setter
    def qubit_variables(self, value: builtins.list[builtins.str]) -> None: ...
    def __eq__(self, other:builtins.object) -> builtins.bool: ...
    def __getnewargs__(self) -> tuple[builtins.str, builtins.list[builtins.str], builtins.list[builtins.str], builtins.list[Instruction]]: ...
    def __new__(cls, name:builtins.str, parameters:typing.Sequence[builtins.str], qubit_variables:typing.Sequence[builtins.str], instructions:typing.Sequence[Instruction]) -> CircuitDefinition: ...
    def __repr__(self) -> builtins.str: ...
    def to_quil(self) -> builtins.str: ...
    def to_quil_or_debug(self) -> builtins.str: ...

class Comparison:
    @property
    def destination(self) -> MemoryReference: ...
    @property
    def lhs(self) -> MemoryReference: ...
    @property
    def operator(self) -> ComparisonOperator: ...
    @property
    def rhs(self) -> ComparisonOperand: ...
    def __eq__(self, other:builtins.object) -> builtins.bool: ...
    def __getnewargs__(self) -> tuple[ComparisonOperator, MemoryReference, MemoryReference, ComparisonOperand]: ...
    def __hash__(self) -> builtins.int: ...
    def __new__(cls, operator:ComparisonOperator, destination:MemoryReference, lhs:MemoryReference, rhs:ComparisonOperand) -> Comparison: ...
    def __repr__(self) -> builtins.str: ...
    def to_quil(self) -> builtins.str: ...
    def to_quil_or_debug(self) -> builtins.str: ...

class ComparisonOperand:
    def __getnewargs__(self) -> tuple[int | float | MemoryReference]: ...
    def __repr__(self) -> builtins.str: ...
    def to_quil(self) -> builtins.str: ...
    def to_quil_or_debug(self) -> builtins.str: ...
    class LiteralInteger(ComparisonOperand):
        __match_args__ = ("_0",)
        @property
        def _0(self) -> builtins.int: ...
        def __getitem__(self, key:builtins.int) -> typing.Any: ...
        def __len__(self) -> builtins.int: ...
        def __new__(cls, _0:builtins.int) -> ComparisonOperand.LiteralInteger: ...
    
    class LiteralReal(ComparisonOperand):
        __match_args__ = ("_0",)
        @property
        def _0(self) -> builtins.float: ...
        def __getitem__(self, key:builtins.int) -> typing.Any: ...
        def __len__(self) -> builtins.int: ...
        def __new__(cls, _0:builtins.float) -> ComparisonOperand.LiteralReal: ...
    
    class MemoryReference(ComparisonOperand):
        __match_args__ = ("_0",)
        @property
        def _0(self) -> MemoryReference: ...
        def __getitem__(self, key:builtins.int) -> typing.Any: ...
        def __len__(self) -> builtins.int: ...
        def __new__(cls, _0:MemoryReference) -> ComparisonOperand.MemoryReference: ...
    

class Convert:
    @property
    def destination(self) -> MemoryReference: ...
    @property
    def source(self) -> MemoryReference: ...
    def __eq__(self, other:builtins.object) -> builtins.bool: ...
    def __getnewargs__(self) -> tuple[MemoryReference, MemoryReference]: ...
    def __hash__(self) -> builtins.int: ...
    def __new__(cls, destination:MemoryReference, source:MemoryReference) -> Convert: ...
    def __repr__(self) -> builtins.str: ...
    def to_quil(self) -> builtins.str: ...
    def to_quil_or_debug(self) -> builtins.str: ...

class Declaration:
    @property
    def name(self) -> builtins.str: ...
    @property
    def sharing(self) -> typing.Optional[Sharing]: ...
    @property
    def size(self) -> Vector: ...
    def __eq__(self, other:builtins.object) -> builtins.bool: ...
    def __getnewargs__(self) -> tuple[builtins.str, Vector, typing.Optional[Sharing]]: ...
    def __hash__(self) -> builtins.int: ...
    def __new__(cls, name:builtins.str, size:Vector, sharing:typing.Optional[Sharing]) -> Declaration: ...
    def __repr__(self) -> builtins.str: ...
    def to_quil(self) -> builtins.str: ...
    def to_quil_or_debug(self) -> builtins.str: ...

class Delay:
    @property
    def duration(self) -> Expression: ...
    @property
    def frame_names(self) -> builtins.list[builtins.str]: ...
    @property
    def qubits(self) -> builtins.list[Qubit]: ...
    def __eq__(self, other:builtins.object) -> builtins.bool: ...
    def __getnewargs__(self) -> tuple[Expression, builtins.list[builtins.str], builtins.list[Qubit]]: ...
    def __hash__(self) -> builtins.int: ...
    def __new__(cls, duration:Expression, frame_names:typing.Sequence[builtins.str], qubits:typing.Sequence[Qubit]) -> Delay: ...
    def __repr__(self) -> builtins.str: ...
    def to_quil(self) -> builtins.str: ...
    def to_quil_or_debug(self) -> builtins.str: ...

class Exchange:
    @property
    def left(self) -> MemoryReference: ...
    @property
    def right(self) -> MemoryReference: ...
    def __eq__(self, other:builtins.object) -> builtins.bool: ...
    def __getnewargs__(self) -> tuple[MemoryReference, MemoryReference]: ...
    def __hash__(self) -> builtins.int: ...
    def __new__(cls, left:MemoryReference, right:MemoryReference) -> Exchange: ...
    def __repr__(self) -> builtins.str: ...
    def to_quil(self) -> builtins.str: ...
    def to_quil_or_debug(self) -> builtins.str: ...

class ExternError(builtins.QuilError):
    r"""
    Errors that may occur when initializing or validating a ``PRAGMA EXTERN`` instruction.
    """
    ...

class ExternParameter:
    r"""
    An extern parameter with a name, mutability, and data type.
    """
    @property
    def data_type(self) -> ExternParameterType:
        r"""
        The data type of the parameter.
        """
    @property
    def mutable(self) -> builtins.bool:
        r"""
        Whether the parameter is mutable.
        """
    @property
    def name(self) -> builtins.str:
        r"""
        The name of the parameter. This must be a valid user identifier.
        """
    def __eq__(self, other:builtins.object) -> builtins.bool: ...
    def __getnewargs__(self) -> tuple[builtins.str, builtins.bool, ExternParameterType]: ...
    def __hash__(self) -> builtins.int: ...
    def __new__(cls, name:builtins.str, mutable:builtins.bool, data_type:ExternParameterType) -> ExternParameter:
        r"""
        Create a new extern parameter. This will fail if the parameter name
        is not a valid user identifier.
        """
    def __repr__(self) -> builtins.str: ...
    def to_quil(self) -> builtins.str: ...
    def to_quil_or_debug(self) -> builtins.str: ...

class ExternParameterType:
    r"""
    A parameter type within an extern signature.
    """
    def __getnewargs__(self) -> tuple[ScalarType | Vector]: ...
    def __repr__(self) -> builtins.str: ...
    def to_quil(self) -> builtins.str: ...
    def to_quil_or_debug(self) -> builtins.str: ...
    class FixedLengthVector(ExternParameterType):
        r"""
        A fixed-length vector, which must accept a memory region name of the appropriate
        length and data type.
        
        For instance `PRAGMA EXTERN foo "(bar : INTEGER[2])"`.
        """
        __match_args__ = ("_0",)
        @property
        def _0(self) -> Vector: ...
        def __getitem__(self, key:builtins.int) -> typing.Any: ...
        def __len__(self) -> builtins.int: ...
        def __new__(cls, _0:Vector) -> ExternParameterType.FixedLengthVector: ...
    
    class Scalar(ExternParameterType):
        r"""
        A scalar parameter, which may accept a memory reference or immediate value.
        
        For instance `PRAGMA EXTERN foo "(bar : INTEGER)"`.
        """
        __match_args__ = ("_0",)
        @property
        def _0(self) -> ScalarType: ...
        def __getitem__(self, key:builtins.int) -> typing.Any: ...
        def __len__(self) -> builtins.int: ...
        def __new__(cls, _0:ScalarType) -> ExternParameterType.Scalar: ...
    
    class VariableLengthVector(ExternParameterType):
        r"""
        A variable-length vector, which must accept a memory region name of the appropriate
        data type.
        
        For instance `PRAGMA EXTERN foo "(bar : INTEGER[])"`.
        """
        __match_args__ = ("_0",)
        @property
        def _0(self) -> ScalarType: ...
        def __getitem__(self, key:builtins.int) -> typing.Any: ...
        def __len__(self) -> builtins.int: ...
        def __new__(cls, _0:ScalarType) -> ExternParameterType.VariableLengthVector: ...
    

class ExternSignature:
    r"""
    An extern signature with a return type and parameters.
    
    The signature of a ``PRAGMA EXTERN`` instruction.
    This signature is defined by a list of ``ExternParameter``s and an optional return type.
    See the [Quil Specification](https://github.com/quil-lang/quil/blob/7f532c7cdde9f51eae6abe7408cc868fba9f91f6/specgen/spec/sec-other.s)
    for details on how these signatures are formed.
    """
    @property
    def parameters(self) -> builtins.list[ExternParameter]:
        r"""
        The parameters of the extern signature.
        """
    @property
    def return_type(self) -> typing.Optional[ScalarType]:
        r"""
        The return type of the extern signature, if any.
        """
    def __eq__(self, other:builtins.object) -> builtins.bool: ...
    def __getnewargs__(self) -> tuple[typing.Optional[ScalarType], builtins.list[ExternParameter]]: ...
    def __hash__(self) -> builtins.int: ...
    def __new__(cls, return_type:typing.Optional[ScalarType], parameters:typing.Sequence[ExternParameter]) -> ExternSignature:
        r"""
        Create a new extern signature.
        """
    def __repr__(self) -> builtins.str: ...
    def to_quil(self) -> builtins.str: ...
    def to_quil_or_debug(self) -> builtins.str: ...

class Fence:
    @property
    def qubits(self) -> builtins.list[Qubit]: ...
    def __eq__(self, other:builtins.object) -> builtins.bool: ...
    def __getnewargs__(self) -> tuple[builtins.list[Qubit]]: ...
    def __hash__(self) -> builtins.int: ...
    def __new__(cls, qubits:typing.Sequence[Qubit]) -> Fence: ...
    def __repr__(self) -> builtins.str: ...
    def to_quil(self) -> builtins.str: ...
    def to_quil_or_debug(self) -> builtins.str: ...

class FrameDefinition:
    @property
    def attributes(self) -> builtins.dict[builtins.str, AttributeValue]: ...
    @property
    def identifier(self) -> FrameIdentifier: ...
    @attributes.setter
    def attributes(self, value: builtins.dict[builtins.str, AttributeValue]) -> None: ...
    @identifier.setter
    def identifier(self, value: FrameIdentifier) -> None: ...
    def __eq__(self, other:builtins.object) -> builtins.bool: ...
    def __getnewargs__(self) -> tuple[FrameIdentifier, builtins.dict[builtins.str, AttributeValue]]: ...
    def __new__(cls, identifier:FrameIdentifier, attributes:typing.Mapping[builtins.str, AttributeValue]) -> FrameDefinition: ...
    def __repr__(self) -> builtins.str: ...
    def to_quil(self) -> builtins.str: ...
    def to_quil_or_debug(self) -> builtins.str: ...

class FrameIdentifier:
    @property
    def name(self) -> builtins.str: ...
    @property
    def qubits(self) -> builtins.list[Qubit]: ...
    def __eq__(self, other:builtins.object) -> builtins.bool: ...
    def __getnewargs__(self) -> tuple[builtins.str, builtins.list[Qubit]]: ...
    def __hash__(self) -> builtins.int: ...
    def __new__(cls, name:builtins.str, qubits:typing.Sequence[Qubit]) -> FrameIdentifier: ...
    def __repr__(self) -> builtins.str: ...
    def to_quil(self) -> builtins.str: ...
    def to_quil_or_debug(self) -> builtins.str: ...

class Gate:
    r"""
    A struct encapsulating all the properties of a Quil Quantum Gate.
    """
    @property
    def modifiers(self) -> builtins.list[GateModifier]: ...
    @property
    def name(self) -> builtins.str: ...
    @property
    def parameters(self) -> builtins.list[Expression]: ...
    @property
    def qubits(self) -> builtins.list[Qubit]: ...
    def __eq__(self, other:builtins.object) -> builtins.bool: ...
    def __getnewargs__(self) -> tuple[builtins.str, builtins.list[Expression], builtins.list[Qubit], builtins.list[GateModifier]]: ...
    def __hash__(self) -> builtins.int: ...
    def __new__(cls, name:builtins.str, parameters:typing.Sequence[Expression], qubits:typing.Sequence[Qubit], modifiers:typing.Sequence[GateModifier]) -> Gate: ...
    def __repr__(self) -> builtins.str: ...
    def controlled(self, control_qubit:Qubit) -> Gate:
        r"""
        Return a copy of the ``Gate`` with the ``CONTROLLED`` modifier added to it.
        """
    def dagger(self) -> Gate:
        r"""
        Return a copy of the ``Gate`` with the ``DAGGER`` modifier added to it.
        """
    def forked(self, fork_qubit:Qubit, alt_params:typing.Sequence[Expression]) -> Gate:
        r"""
        Return a copy of the ``Gate`` with the ``FORKED`` modifier added to it.
        
        Raises a ``GateError`` if the number of provided alternate parameters
        don't equal the number of existing parameters.
        """
    def to_quil(self) -> builtins.str: ...
    def to_quil_or_debug(self) -> builtins.str: ...
    def to_unitary(self, n_qubits:builtins.int) -> numpy.typing.NDArray[numpy.complex128]:
        r"""
        Get the matrix resulting from lifting this ``Gate``
        to the full `n_qubits`-qubit Hilbert space.
        
        Raises a ``GateError`` if any of the parameters of this ``Gate`` are non-constant,
        if any of the ``Qubit``s are variable,
        if the name of this ``Gate`` is unknown,
        or if there are an unexpected number of parameters.
        
        # Notes
        
        A previous version of this library called this `to_unitary_mut`,
        and modified the ``Gate`` when called.
        This is no longer possible, as it would modify the ``Gate``'s hash,
        leading to confusing bugs.
        ``Gate``s, as well as all other hashable classes, are immutable from Python.
        
        # Bugs
        
        Supplying `n_qubits` as `0` will raise an unspecified exception;
        other invalid input parameters may silently return an invalid result.
        """

class GateDefinition:
    r"""
    A struct encapsulating a quil Gate Definition
    """
    @property
    def name(self) -> builtins.str: ...
    @property
    def parameters(self) -> builtins.list[builtins.str]: ...
    @property
    def specification(self) -> GateSpecification: ...
    def __eq__(self, other:builtins.object) -> builtins.bool: ...
    def __getnewargs__(self) -> tuple[builtins.str, builtins.list[builtins.str], GateSpecification]: ...
    def __hash__(self) -> builtins.int: ...
    def __new__(cls, name:builtins.str, parameters:typing.Sequence[builtins.str], specification:GateSpecification) -> GateDefinition: ...
    def __repr__(self) -> builtins.str: ...
    def to_quil(self) -> builtins.str: ...
    def to_quil_or_debug(self) -> builtins.str: ...

class GateError(builtins.QuilError):
    r"""
    Errors that may occur when performing operations on a ``Gate``.
    """
    ...

class GateSpecification:
    r"""
    An enum representing a the specification of a [`GateDefinition`] for a given [`GateType`]
    """
    def __getnewargs__(self) -> tuple[list[list[Expression]] | list[int] | GateType.PauliSum]: ...
    def __repr__(self) -> builtins.str: ...
    def to_quil(self) -> builtins.str: ...
    def to_quil_or_debug(self) -> builtins.str: ...
    class Matrix(GateSpecification):
        r"""
        A matrix of [`Expression`]s representing a unitary operation for a [`GateType::Matrix`].
        """
        __match_args__ = ("_0",)
        @property
        def _0(self) -> builtins.list[builtins.list[Expression]]: ...
        def __getitem__(self, key:builtins.int) -> typing.Any: ...
        def __len__(self) -> builtins.int: ...
        def __new__(cls, _0:typing.Sequence[typing.Sequence[Expression]]) -> GateSpecification.Matrix: ...
    
    class PauliSum(GateSpecification):
        r"""
        A Hermitian operator specified as a Pauli sum, a sum of combinations of Pauli operators,
        used for a [`GateType::PauliSum`]
        """
        __match_args__ = ("_0",)
        @property
        def _0(self) -> PauliSum: ...
        def __getitem__(self, key:builtins.int) -> typing.Any: ...
        def __len__(self) -> builtins.int: ...
        def __new__(cls, _0:PauliSum) -> GateSpecification.PauliSum: ...
    
    class Permutation(GateSpecification):
        r"""
        A vector of integers that defines the permutation used for a [`GateType::Permutation`]
        """
        __match_args__ = ("_0",)
        @property
        def _0(self) -> builtins.list[builtins.int]: ...
        def __getitem__(self, key:builtins.int) -> typing.Any: ...
        def __len__(self) -> builtins.int: ...
        def __new__(cls, _0:typing.Sequence[builtins.int]) -> GateSpecification.Permutation: ...
    

class Include:
    @property
    def filename(self) -> builtins.str: ...
    def __eq__(self, other:builtins.object) -> builtins.bool: ...
    def __getnewargs__(self) -> tuple[builtins.str]: ...
    def __hash__(self) -> builtins.int: ...
    def __new__(cls, filename:builtins.str) -> Include: ...
    def __repr__(self) -> builtins.str: ...
    def to_quil(self) -> builtins.str: ...
    def to_quil_or_debug(self) -> builtins.str: ...

class Instruction:
    r"""
    A Quil instruction.
    
    Each variant (for Python users, each nested subclass)
    corresponds to a possible type of Quil instruction,
    which is accessible as a member within the variant.
    
    # Python Users
    
    The subclasses of this class are class attributes defined on it,
    and can be used to "wrap" instructions when they should be stored together.
    In particular, they are *NOT* the instruction classes you'd typically create,
    and instances of instruction classes are *NOT* subclasses of this class:
    
    ```python
    >>> from quil.instructions import Instruction, Gate, Qubit
    >>> issubclass(Instruction.Gate, Instruction)
    True
    >>> issubclass(Gate, Instruction)
    False
    >>> g = Gate("X", (), (Qubit.Fixed(0),), ())
    >>> isinstance(g, Gate)
    True
    >>> isinstance(g, Instruction.Gate)
    False
    >>> g_instr = Instruction.Gate(g)
    >>> isinstance(g_instr, Gate)
    False
    >>> isinstance(g_instr, Instruction.Gate)
    True
    >>> isinstance(g_instr._0, Gate)
    True
    >>> g_instr._0 == g
    True
    ```
    
    The point of this class is to wrap different kinds of instructions
    when stored together in a collection, all of which are of type `Instruction`.
    You can check for different instruction variants and destructure them using `match`:
    
    ```python
    match g_instr:
        case Instruction.Gate(gate):
            assert isinstance(gate, Gate)
        case Instruction.Wait() | Instruction.Nop():
            # note the `()` -- these aren't like Python's enumerations!
    ```
    """
    def __getnewargs__(self) -> tuple[()] | tuple[Arithmetic | BinaryLogic | Call | Capture | CalibrationDefinition | CircuitDefinition | Comparison | Convert | Declaration | Delay | Exchange | Fence | FrameDefinition | Gate | GateDefinition | Include | Jump | JumpUnless | JumpWhen | Label | Load | MeasureCalibrationDefinition | Measurement | Move | Pragma | Pulse | RawCapture | Reset | SetFrequency | SetPhase | SetScale | ShiftFrequency | ShiftPhase | Store | SwapPhases | UnaryLogic | WaveformDefinition]: ...
    def __repr__(self) -> builtins.str: ...
    def is_quil_t(self) -> builtins.bool:
        r"""
        Returns true if the instruction is a Quil-T instruction.
        """
    @staticmethod
    def parse(string:builtins.str) -> Instruction: ...
    def to_quil(self) -> builtins.str: ...
    def to_quil_or_debug(self) -> builtins.str: ...
    class Arithmetic(Instruction):
        __match_args__ = ("_0",)
        @property
        def _0(self) -> Arithmetic: ...
        def __getitem__(self, key:builtins.int) -> typing.Any: ...
        def __len__(self) -> builtins.int: ...
        def __new__(cls, _0:Arithmetic) -> Instruction.Arithmetic: ...
    
    class BinaryLogic(Instruction):
        __match_args__ = ("_0",)
        @property
        def _0(self) -> BinaryLogic: ...
        def __getitem__(self, key:builtins.int) -> typing.Any: ...
        def __len__(self) -> builtins.int: ...
        def __new__(cls, _0:BinaryLogic) -> Instruction.BinaryLogic: ...
    
    class CalibrationDefinition(Instruction):
        __match_args__ = ("_0",)
        @property
        def _0(self) -> CalibrationDefinition: ...
        def __getitem__(self, key:builtins.int) -> typing.Any: ...
        def __len__(self) -> builtins.int: ...
        def __new__(cls, _0:CalibrationDefinition) -> Instruction.CalibrationDefinition: ...
    
    class Call(Instruction):
        __match_args__ = ("_0",)
        @property
        def _0(self) -> Call: ...
        def __getitem__(self, key:builtins.int) -> typing.Any: ...
        def __len__(self) -> builtins.int: ...
        def __new__(cls, _0:Call) -> Instruction.Call: ...
    
    class Capture(Instruction):
        __match_args__ = ("_0",)
        @property
        def _0(self) -> Capture: ...
        def __getitem__(self, key:builtins.int) -> typing.Any: ...
        def __len__(self) -> builtins.int: ...
        def __new__(cls, _0:Capture) -> Instruction.Capture: ...
    
    class CircuitDefinition(Instruction):
        __match_args__ = ("_0",)
        @property
        def _0(self) -> CircuitDefinition: ...
        def __getitem__(self, key:builtins.int) -> typing.Any: ...
        def __len__(self) -> builtins.int: ...
        def __new__(cls, _0:CircuitDefinition) -> Instruction.CircuitDefinition: ...
    
    class Comparison(Instruction):
        __match_args__ = ("_0",)
        @property
        def _0(self) -> Comparison: ...
        def __getitem__(self, key:builtins.int) -> typing.Any: ...
        def __len__(self) -> builtins.int: ...
        def __new__(cls, _0:Comparison) -> Instruction.Comparison: ...
    
    class Convert(Instruction):
        __match_args__ = ("_0",)
        @property
        def _0(self) -> Convert: ...
        def __getitem__(self, key:builtins.int) -> typing.Any: ...
        def __len__(self) -> builtins.int: ...
        def __new__(cls, _0:Convert) -> Instruction.Convert: ...
    
    class Declaration(Instruction):
        __match_args__ = ("_0",)
        @property
        def _0(self) -> Declaration: ...
        def __getitem__(self, key:builtins.int) -> typing.Any: ...
        def __len__(self) -> builtins.int: ...
        def __new__(cls, _0:Declaration) -> Instruction.Declaration: ...
    
    class Delay(Instruction):
        __match_args__ = ("_0",)
        @property
        def _0(self) -> Delay: ...
        def __getitem__(self, key:builtins.int) -> typing.Any: ...
        def __len__(self) -> builtins.int: ...
        def __new__(cls, _0:Delay) -> Instruction.Delay: ...
    
    class Exchange(Instruction):
        __match_args__ = ("_0",)
        @property
        def _0(self) -> Exchange: ...
        def __getitem__(self, key:builtins.int) -> typing.Any: ...
        def __len__(self) -> builtins.int: ...
        def __new__(cls, _0:Exchange) -> Instruction.Exchange: ...
    
    class Fence(Instruction):
        __match_args__ = ("_0",)
        @property
        def _0(self) -> Fence: ...
        def __getitem__(self, key:builtins.int) -> typing.Any: ...
        def __len__(self) -> builtins.int: ...
        def __new__(cls, _0:Fence) -> Instruction.Fence: ...
    
    class FrameDefinition(Instruction):
        __match_args__ = ("_0",)
        @property
        def _0(self) -> FrameDefinition: ...
        def __getitem__(self, key:builtins.int) -> typing.Any: ...
        def __len__(self) -> builtins.int: ...
        def __new__(cls, _0:FrameDefinition) -> Instruction.FrameDefinition: ...
    
    class Gate(Instruction):
        __match_args__ = ("_0",)
        @property
        def _0(self) -> Gate: ...
        def __getitem__(self, key:builtins.int) -> typing.Any: ...
        def __len__(self) -> builtins.int: ...
        def __new__(cls, _0:Gate) -> Instruction.Gate: ...
    
    class GateDefinition(Instruction):
        __match_args__ = ("_0",)
        @property
        def _0(self) -> GateDefinition: ...
        def __getitem__(self, key:builtins.int) -> typing.Any: ...
        def __len__(self) -> builtins.int: ...
        def __new__(cls, _0:GateDefinition) -> Instruction.GateDefinition: ...
    
    class Halt(Instruction):
        __match_args__ = ((),)
        def __getitem__(self, key:builtins.int) -> typing.Any: ...
        def __len__(self) -> builtins.int: ...
        def __new__(cls) -> Instruction.Halt: ...
    
    class Include(Instruction):
        __match_args__ = ("_0",)
        @property
        def _0(self) -> Include: ...
        def __getitem__(self, key:builtins.int) -> typing.Any: ...
        def __len__(self) -> builtins.int: ...
        def __new__(cls, _0:Include) -> Instruction.Include: ...
    
    class Jump(Instruction):
        __match_args__ = ("_0",)
        @property
        def _0(self) -> Jump: ...
        def __getitem__(self, key:builtins.int) -> typing.Any: ...
        def __len__(self) -> builtins.int: ...
        def __new__(cls, _0:Jump) -> Instruction.Jump: ...
    
    class JumpUnless(Instruction):
        __match_args__ = ("_0",)
        @property
        def _0(self) -> JumpUnless: ...
        def __getitem__(self, key:builtins.int) -> typing.Any: ...
        def __len__(self) -> builtins.int: ...
        def __new__(cls, _0:JumpUnless) -> Instruction.JumpUnless: ...
    
    class JumpWhen(Instruction):
        __match_args__ = ("_0",)
        @property
        def _0(self) -> JumpWhen: ...
        def __getitem__(self, key:builtins.int) -> typing.Any: ...
        def __len__(self) -> builtins.int: ...
        def __new__(cls, _0:JumpWhen) -> Instruction.JumpWhen: ...
    
    class Label(Instruction):
        __match_args__ = ("_0",)
        @property
        def _0(self) -> Label: ...
        def __getitem__(self, key:builtins.int) -> typing.Any: ...
        def __len__(self) -> builtins.int: ...
        def __new__(cls, _0:Label) -> Instruction.Label: ...
    
    class Load(Instruction):
        __match_args__ = ("_0",)
        @property
        def _0(self) -> Load: ...
        def __getitem__(self, key:builtins.int) -> typing.Any: ...
        def __len__(self) -> builtins.int: ...
        def __new__(cls, _0:Load) -> Instruction.Load: ...
    
    class MeasureCalibrationDefinition(Instruction):
        __match_args__ = ("_0",)
        @property
        def _0(self) -> MeasureCalibrationDefinition: ...
        def __getitem__(self, key:builtins.int) -> typing.Any: ...
        def __len__(self) -> builtins.int: ...
        def __new__(cls, _0:MeasureCalibrationDefinition) -> Instruction.MeasureCalibrationDefinition: ...
    
    class Measurement(Instruction):
        __match_args__ = ("_0",)
        @property
        def _0(self) -> Measurement: ...
        def __getitem__(self, key:builtins.int) -> typing.Any: ...
        def __len__(self) -> builtins.int: ...
        def __new__(cls, _0:Measurement) -> Instruction.Measurement: ...
    
    class Move(Instruction):
        __match_args__ = ("_0",)
        @property
        def _0(self) -> Move: ...
        def __getitem__(self, key:builtins.int) -> typing.Any: ...
        def __len__(self) -> builtins.int: ...
        def __new__(cls, _0:Move) -> Instruction.Move: ...
    
    class Nop(Instruction):
        __match_args__ = ((),)
        def __getitem__(self, key:builtins.int) -> typing.Any: ...
        def __len__(self) -> builtins.int: ...
        def __new__(cls) -> Instruction.Nop: ...
    
    class Pragma(Instruction):
        __match_args__ = ("_0",)
        @property
        def _0(self) -> Pragma: ...
        def __getitem__(self, key:builtins.int) -> typing.Any: ...
        def __len__(self) -> builtins.int: ...
        def __new__(cls, _0:Pragma) -> Instruction.Pragma: ...
    
    class Pulse(Instruction):
        __match_args__ = ("_0",)
        @property
        def _0(self) -> Pulse: ...
        def __getitem__(self, key:builtins.int) -> typing.Any: ...
        def __len__(self) -> builtins.int: ...
        def __new__(cls, _0:Pulse) -> Instruction.Pulse: ...
    
    class RawCapture(Instruction):
        __match_args__ = ("_0",)
        @property
        def _0(self) -> RawCapture: ...
        def __getitem__(self, key:builtins.int) -> typing.Any: ...
        def __len__(self) -> builtins.int: ...
        def __new__(cls, _0:RawCapture) -> Instruction.RawCapture: ...
    
    class Reset(Instruction):
        __match_args__ = ("_0",)
        @property
        def _0(self) -> Reset: ...
        def __getitem__(self, key:builtins.int) -> typing.Any: ...
        def __len__(self) -> builtins.int: ...
        def __new__(cls, _0:Reset) -> Instruction.Reset: ...
    
    class SetFrequency(Instruction):
        __match_args__ = ("_0",)
        @property
        def _0(self) -> SetFrequency: ...
        def __getitem__(self, key:builtins.int) -> typing.Any: ...
        def __len__(self) -> builtins.int: ...
        def __new__(cls, _0:SetFrequency) -> Instruction.SetFrequency: ...
    
    class SetPhase(Instruction):
        __match_args__ = ("_0",)
        @property
        def _0(self) -> SetPhase: ...
        def __getitem__(self, key:builtins.int) -> typing.Any: ...
        def __len__(self) -> builtins.int: ...
        def __new__(cls, _0:SetPhase) -> Instruction.SetPhase: ...
    
    class SetScale(Instruction):
        __match_args__ = ("_0",)
        @property
        def _0(self) -> SetScale: ...
        def __getitem__(self, key:builtins.int) -> typing.Any: ...
        def __len__(self) -> builtins.int: ...
        def __new__(cls, _0:SetScale) -> Instruction.SetScale: ...
    
    class ShiftFrequency(Instruction):
        __match_args__ = ("_0",)
        @property
        def _0(self) -> ShiftFrequency: ...
        def __getitem__(self, key:builtins.int) -> typing.Any: ...
        def __len__(self) -> builtins.int: ...
        def __new__(cls, _0:ShiftFrequency) -> Instruction.ShiftFrequency: ...
    
    class ShiftPhase(Instruction):
        __match_args__ = ("_0",)
        @property
        def _0(self) -> ShiftPhase: ...
        def __getitem__(self, key:builtins.int) -> typing.Any: ...
        def __len__(self) -> builtins.int: ...
        def __new__(cls, _0:ShiftPhase) -> Instruction.ShiftPhase: ...
    
    class Store(Instruction):
        __match_args__ = ("_0",)
        @property
        def _0(self) -> Store: ...
        def __getitem__(self, key:builtins.int) -> typing.Any: ...
        def __len__(self) -> builtins.int: ...
        def __new__(cls, _0:Store) -> Instruction.Store: ...
    
    class SwapPhases(Instruction):
        __match_args__ = ("_0",)
        @property
        def _0(self) -> SwapPhases: ...
        def __getitem__(self, key:builtins.int) -> typing.Any: ...
        def __len__(self) -> builtins.int: ...
        def __new__(cls, _0:SwapPhases) -> Instruction.SwapPhases: ...
    
    class UnaryLogic(Instruction):
        __match_args__ = ("_0",)
        @property
        def _0(self) -> UnaryLogic: ...
        def __getitem__(self, key:builtins.int) -> typing.Any: ...
        def __len__(self) -> builtins.int: ...
        def __new__(cls, _0:UnaryLogic) -> Instruction.UnaryLogic: ...
    
    class Wait(Instruction):
        __match_args__ = ((),)
        def __getitem__(self, key:builtins.int) -> typing.Any: ...
        def __len__(self) -> builtins.int: ...
        def __new__(cls) -> Instruction.Wait: ...
    
    class WaveformDefinition(Instruction):
        __match_args__ = ("_0",)
        @property
        def _0(self) -> WaveformDefinition: ...
        def __getitem__(self, key:builtins.int) -> typing.Any: ...
        def __len__(self) -> builtins.int: ...
        def __new__(cls, _0:WaveformDefinition) -> Instruction.WaveformDefinition: ...
    

class InstructionError(builtins.QuilError):
    r"""
    Base error type for errors related to ``Instruction`` processing.
    """
    ...

class Jump:
    @property
    def target(self) -> Target: ...
    @target.setter
    def target(self, value: Target) -> None: ...
    def __eq__(self, other:builtins.object) -> builtins.bool: ...
    def __getnewargs__(self) -> tuple[Target]: ...
    def __new__(cls, target:Target) -> Jump: ...
    def __repr__(self) -> builtins.str: ...
    def to_quil(self) -> builtins.str: ...
    def to_quil_or_debug(self) -> builtins.str: ...

class JumpUnless:
    @property
    def condition(self) -> MemoryReference: ...
    @property
    def target(self) -> Target: ...
    @condition.setter
    def condition(self, value: MemoryReference) -> None: ...
    @target.setter
    def target(self, value: Target) -> None: ...
    def __eq__(self, other:builtins.object) -> builtins.bool: ...
    def __getnewargs__(self) -> tuple[Target, MemoryReference]: ...
    def __new__(cls, target:Target, condition:MemoryReference) -> JumpUnless: ...
    def __repr__(self) -> builtins.str: ...
    def to_quil(self) -> builtins.str: ...
    def to_quil_or_debug(self) -> builtins.str: ...

class JumpWhen:
    @property
    def condition(self) -> MemoryReference: ...
    @property
    def target(self) -> Target: ...
    @condition.setter
    def condition(self, value: MemoryReference) -> None: ...
    @target.setter
    def target(self, value: Target) -> None: ...
    def __eq__(self, other:builtins.object) -> builtins.bool: ...
    def __getnewargs__(self) -> tuple[Target, MemoryReference]: ...
    def __new__(cls, target:Target, condition:MemoryReference) -> JumpWhen: ...
    def __repr__(self) -> builtins.str: ...
    def to_quil(self) -> builtins.str: ...
    def to_quil_or_debug(self) -> builtins.str: ...

class Label:
    @property
    def target(self) -> Target: ...
    def __eq__(self, other:builtins.object) -> builtins.bool: ...
    def __getnewargs__(self) -> tuple[Target]: ...
    def __hash__(self) -> builtins.int: ...
    def __new__(cls, target:Target) -> Label: ...
    def __repr__(self) -> builtins.str: ...
    def to_quil(self) -> builtins.str: ...
    def to_quil_or_debug(self) -> builtins.str: ...

class Load:
    @property
    def destination(self) -> MemoryReference: ...
    @property
    def offset(self) -> MemoryReference: ...
    @property
    def source(self) -> builtins.str: ...
    def __eq__(self, other:builtins.object) -> builtins.bool: ...
    def __getnewargs__(self) -> tuple[MemoryReference, builtins.str, MemoryReference]: ...
    def __hash__(self) -> builtins.int: ...
    def __new__(cls, destination:MemoryReference, source:builtins.str, offset:MemoryReference) -> Load: ...
    def __repr__(self) -> builtins.str: ...
    def to_quil(self) -> builtins.str: ...
    def to_quil_or_debug(self) -> builtins.str: ...

class MeasureCalibrationDefinition:
    @property
    def identifier(self) -> MeasureCalibrationIdentifier: ...
    @property
    def instructions(self) -> builtins.list[Instruction]: ...
    @property
    def qubit(self) -> Qubit:
        r"""
        The qubit that this measure calibration definition is for.
        """
    @property
    def target(self) -> typing.Optional[builtins.str]:
        r"""
        The name the measurement calibration uses for the variable it will write the measurement
        result to, if this is a measurement for record.
        """
    @identifier.setter
    def identifier(self, value: MeasureCalibrationIdentifier) -> None: ...
    @instructions.setter
    def instructions(self, value: builtins.list[Instruction]) -> None: ...
    def __eq__(self, other:builtins.object) -> builtins.bool: ...
    def __getnewargs__(self) -> tuple[MeasureCalibrationIdentifier, builtins.list[Instruction]]: ...
    def __new__(cls, identifier:MeasureCalibrationIdentifier, instructions:typing.Sequence[Instruction]) -> MeasureCalibrationDefinition: ...
    def __repr__(self) -> builtins.str: ...
    def to_quil(self) -> builtins.str: ...
    def to_quil_or_debug(self) -> builtins.str: ...

class MeasureCalibrationIdentifier:
    r"""
    A unique identifier for a measurement calibration definition within a program
    """
    @property
    def qubit(self) -> Qubit:
        r"""
        The qubit which is being measured.
        """
    @property
    def target(self) -> typing.Optional[builtins.str]:
        r"""
        The name the definition uses for the variable it will write the measurement result to, if
        this is a measurement for record.
        
        If this is missing, this is a calibration for a measurement for effect.
        """
    @qubit.setter
    def qubit(self, value: Qubit) -> None:
        r"""
        The qubit which is being measured.
        """
    @target.setter
    def target(self, value: typing.Optional[builtins.str]) -> None:
        r"""
        The name the definition uses for the variable it will write the measurement result to, if
        this is a measurement for record.
        
        If this is missing, this is a calibration for a measurement for effect.
        """
    def __eq__(self, other:builtins.object) -> builtins.bool: ...
    def __getnewargs__(self) -> tuple[Qubit, typing.Optional[builtins.str]]: ...
    def __new__(cls, qubit:Qubit, target:typing.Optional[builtins.str]) -> MeasureCalibrationIdentifier: ...
    def __repr__(self) -> builtins.str: ...
    def to_quil(self) -> builtins.str: ...
    def to_quil_or_debug(self) -> builtins.str: ...

class Measurement:
    @property
    def qubit(self) -> Qubit: ...
    @property
    def target(self) -> typing.Optional[MemoryReference]: ...
    def __eq__(self, other:builtins.object) -> builtins.bool: ...
    def __getnewargs__(self) -> tuple[Qubit, typing.Optional[MemoryReference]]: ...
    def __hash__(self) -> builtins.int: ...
    def __new__(cls, qubit:Qubit, target:typing.Optional[MemoryReference]) -> Measurement: ...
    def __repr__(self) -> builtins.str: ...
    def to_quil(self) -> builtins.str: ...
    def to_quil_or_debug(self) -> builtins.str: ...

class MemoryReference:
    @property
    def index(self) -> builtins.int: ...
    @property
    def name(self) -> builtins.str: ...
    def __eq__(self, other:builtins.object) -> builtins.bool: ...
    def __getnewargs__(self) -> tuple[builtins.str, builtins.int]: ...
    def __hash__(self) -> builtins.int: ...
    def __new__(cls, name:builtins.str, index:builtins.int) -> MemoryReference: ...
    def __repr__(self) -> builtins.str: ...
    @staticmethod
    def parse(string:builtins.str) -> MemoryReference: ...
    def to_quil(self) -> builtins.str: ...
    def to_quil_or_debug(self) -> builtins.str: ...

class Move:
    @property
    def destination(self) -> MemoryReference: ...
    @property
    def source(self) -> ArithmeticOperand: ...
    def __eq__(self, other:builtins.object) -> builtins.bool: ...
    def __getnewargs__(self) -> tuple[MemoryReference, ArithmeticOperand]: ...
    def __hash__(self) -> builtins.int: ...
    def __new__(cls, destination:MemoryReference, source:ArithmeticOperand) -> Move: ...
    def __repr__(self) -> builtins.str: ...
    def to_quil(self) -> builtins.str: ...
    def to_quil_or_debug(self) -> builtins.str: ...

class Offset:
    @property
    def data_type(self) -> ScalarType: ...
    @property
    def offset(self) -> builtins.int: ...
    def __eq__(self, other:builtins.object) -> builtins.bool: ...
    def __getnewargs__(self) -> tuple[builtins.int, ScalarType]: ...
    def __hash__(self) -> builtins.int: ...
    def __new__(cls, offset:builtins.int, data_type:ScalarType) -> Offset: ...
    def __repr__(self) -> builtins.str: ...
    def to_quil(self) -> builtins.str: ...
    def to_quil_or_debug(self) -> builtins.str: ...

class ParseInstructionError(builtins.InstructionError):
    r"""
    Errors that may occur while parsing an ``Instruction``.
    """
    ...

class ParseMemoryReferenceError(builtins.QuilError):
    r"""
    Errors that may occur while parsing a ``MemoryReference``.
    """
    ...

class PauliSum:
    @property
    def arguments(self) -> builtins.list[builtins.str]: ...
    @property
    def terms(self) -> builtins.list[PauliTerm]: ...
    def __eq__(self, other:builtins.object) -> builtins.bool: ...
    def __getnewargs__(self) -> tuple[builtins.list[builtins.str], builtins.list[PauliTerm]]: ...
    def __hash__(self) -> builtins.int: ...
    def __new__(cls, arguments:typing.Sequence[builtins.str], terms:typing.Sequence[PauliTerm]) -> PauliSum: ...
    def __repr__(self) -> builtins.str: ...

class PauliTerm:
    @property
    def arguments(self) -> builtins.list[tuple[PauliGate, builtins.str]]: ...
    @property
    def expression(self) -> Expression: ...
    def __eq__(self, other:builtins.object) -> builtins.bool: ...
    def __getnewargs__(self) -> tuple[builtins.list[tuple[PauliGate, builtins.str]], Expression]: ...
    def __hash__(self) -> builtins.int: ...
    def __new__(cls, arguments:typing.Sequence[tuple[PauliGate, builtins.str]], expression:Expression) -> PauliTerm: ...
    def __repr__(self) -> builtins.str: ...

class Pragma:
    @property
    def arguments(self) -> builtins.list[PragmaArgument]: ...
    @property
    def data(self) -> typing.Optional[builtins.str]: ...
    @property
    def name(self) -> builtins.str: ...
    def __eq__(self, other:builtins.object) -> builtins.bool: ...
    def __getnewargs__(self) -> tuple[builtins.str, builtins.list[PragmaArgument], typing.Optional[builtins.str]]: ...
    def __hash__(self) -> builtins.int: ...
    def __new__(cls, name:builtins.str, arguments:typing.Sequence[PragmaArgument], data:typing.Optional[builtins.str]) -> Pragma: ...
    def __repr__(self) -> builtins.str: ...
    def to_quil(self) -> builtins.str: ...
    def to_quil_or_debug(self) -> builtins.str: ...

class PragmaArgument:
    def __getnewargs__(self) -> tuple[int | str]: ...
    def __repr__(self) -> builtins.str: ...
    def to_quil(self) -> builtins.str: ...
    def to_quil_or_debug(self) -> builtins.str: ...
    class Identifier(PragmaArgument):
        __match_args__ = ("_0",)
        @property
        def _0(self) -> builtins.str: ...
        def __getitem__(self, key:builtins.int) -> typing.Any: ...
        def __len__(self) -> builtins.int: ...
        def __new__(cls, _0:builtins.str) -> PragmaArgument.Identifier: ...
    
    class Integer(PragmaArgument):
        __match_args__ = ("_0",)
        @property
        def _0(self) -> builtins.int: ...
        def __getitem__(self, key:builtins.int) -> typing.Any: ...
        def __len__(self) -> builtins.int: ...
        def __new__(cls, _0:builtins.int) -> PragmaArgument.Integer: ...
    

class Pulse:
    @property
    def blocking(self) -> builtins.bool: ...
    @property
    def frame(self) -> FrameIdentifier: ...
    @property
    def waveform(self) -> WaveformInvocation: ...
    @blocking.setter
    def blocking(self, value: builtins.bool) -> None: ...
    @frame.setter
    def frame(self, value: FrameIdentifier) -> None: ...
    @waveform.setter
    def waveform(self, value: WaveformInvocation) -> None: ...
    def __eq__(self, other:builtins.object) -> builtins.bool: ...
    def __getnewargs__(self) -> tuple[builtins.bool, FrameIdentifier, WaveformInvocation]: ...
    def __new__(cls, blocking:builtins.bool, frame:FrameIdentifier, waveform:WaveformInvocation) -> Pulse: ...
    def __repr__(self) -> builtins.str: ...
    def to_quil(self) -> builtins.str: ...
    def to_quil_or_debug(self) -> builtins.str: ...

class Qubit:
    def __getnewargs__(self) -> tuple[int | str | QubitPlaceholder]: ...
    def __repr__(self) -> builtins.str: ...
    def to_quil(self) -> builtins.str: ...
    def to_quil_or_debug(self) -> builtins.str: ...
    class Fixed(Qubit):
        __match_args__ = ("_0",)
        @property
        def _0(self) -> builtins.int: ...
        def __getitem__(self, key:builtins.int) -> typing.Any: ...
        def __len__(self) -> builtins.int: ...
        def __new__(cls, _0:builtins.int) -> Qubit.Fixed: ...
    
    class Placeholder(Qubit):
        __match_args__ = ("_0",)
        @property
        def _0(self) -> QubitPlaceholder: ...
        def __getitem__(self, key:builtins.int) -> typing.Any: ...
        def __len__(self) -> builtins.int: ...
        def __new__(cls, _0:QubitPlaceholder) -> Qubit.Placeholder: ...
    
    class Variable(Qubit):
        __match_args__ = ("_0",)
        @property
        def _0(self) -> builtins.str: ...
        def __getitem__(self, key:builtins.int) -> typing.Any: ...
        def __len__(self) -> builtins.int: ...
        def __new__(cls, _0:builtins.str) -> Qubit.Variable: ...
    

class QubitPlaceholder:
    r"""
    An opaque placeholder for a qubit whose index may be assigned at a later time.
    """
    def __eq__(self, other:builtins.object) -> builtins.bool: ...
    def __ge__(self, other:builtins.object) -> builtins.bool: ...
    def __getnewargs__(self) -> NoReturn:
        r"""
        `QubitPlaceholder`s do not support `pickle` or `deepcopy`.
        Calling this method will raise an error.
        """
    def __gt__(self, other:builtins.object) -> builtins.bool: ...
    def __hash__(self) -> builtins.int: ...
    def __le__(self, other:builtins.object) -> builtins.bool: ...
    def __lt__(self, other:builtins.object) -> builtins.bool: ...
    def __new__(cls) -> QubitPlaceholder: ...
    def __repr__(self) -> builtins.str: ...

class RawCapture:
    @property
    def blocking(self) -> builtins.bool: ...
    @property
    def duration(self) -> Expression: ...
    @property
    def frame(self) -> FrameIdentifier: ...
    @property
    def memory_reference(self) -> MemoryReference: ...
    def __eq__(self, other:builtins.object) -> builtins.bool: ...
    def __getnewargs__(self) -> tuple[builtins.bool, FrameIdentifier, Expression, MemoryReference]: ...
    def __hash__(self) -> builtins.int: ...
    def __new__(cls, blocking:builtins.bool, frame:FrameIdentifier, duration:Expression, memory_reference:MemoryReference) -> RawCapture: ...
    def __repr__(self) -> builtins.str: ...
    def to_quil(self) -> builtins.str: ...
    def to_quil_or_debug(self) -> builtins.str: ...

class Reset:
    @property
    def qubit(self) -> typing.Optional[Qubit]: ...
    def __eq__(self, other:builtins.object) -> builtins.bool: ...
    def __getnewargs__(self) -> tuple[typing.Optional[Qubit]]: ...
    def __hash__(self) -> builtins.int: ...
    def __new__(cls, qubit:typing.Optional[Qubit]) -> Reset: ...
    def __repr__(self) -> builtins.str: ...
    def to_quil(self) -> builtins.str: ...
    def to_quil_or_debug(self) -> builtins.str: ...

class SetFrequency:
    @property
    def frame(self) -> FrameIdentifier: ...
    @property
    def frequency(self) -> Expression: ...
    def __eq__(self, other:builtins.object) -> builtins.bool: ...
    def __getnewargs__(self) -> tuple[FrameIdentifier, Expression]: ...
    def __hash__(self) -> builtins.int: ...
    def __new__(cls, frame:FrameIdentifier, frequency:Expression) -> SetFrequency: ...
    def __repr__(self) -> builtins.str: ...
    def to_quil(self) -> builtins.str: ...
    def to_quil_or_debug(self) -> builtins.str: ...

class SetPhase:
    @property
    def frame(self) -> FrameIdentifier: ...
    @property
    def phase(self) -> Expression: ...
    def __eq__(self, other:builtins.object) -> builtins.bool: ...
    def __getnewargs__(self) -> tuple[FrameIdentifier, Expression]: ...
    def __hash__(self) -> builtins.int: ...
    def __new__(cls, frame:FrameIdentifier, phase:Expression) -> SetPhase: ...
    def __repr__(self) -> builtins.str: ...
    def to_quil(self) -> builtins.str: ...
    def to_quil_or_debug(self) -> builtins.str: ...

class SetScale:
    @property
    def frame(self) -> FrameIdentifier: ...
    @property
    def scale(self) -> Expression: ...
    def __eq__(self, other:builtins.object) -> builtins.bool: ...
    def __getnewargs__(self) -> tuple[FrameIdentifier, Expression]: ...
    def __hash__(self) -> builtins.int: ...
    def __new__(cls, frame:FrameIdentifier, scale:Expression) -> SetScale: ...
    def __repr__(self) -> builtins.str: ...
    def to_quil(self) -> builtins.str: ...
    def to_quil_or_debug(self) -> builtins.str: ...

class Sharing:
    @property
    def name(self) -> builtins.str: ...
    @property
    def offsets(self) -> builtins.list[Offset]: ...
    def __eq__(self, other:builtins.object) -> builtins.bool: ...
    def __getnewargs__(self) -> tuple[builtins.str, builtins.list[Offset]]: ...
    def __hash__(self) -> builtins.int: ...
    def __new__(cls, name:builtins.str, offsets:typing.Sequence[Offset]) -> Sharing: ...
    def __repr__(self) -> builtins.str: ...

class ShiftFrequency:
    @property
    def frame(self) -> FrameIdentifier: ...
    @property
    def frequency(self) -> Expression: ...
    def __eq__(self, other:builtins.object) -> builtins.bool: ...
    def __getnewargs__(self) -> tuple[FrameIdentifier, Expression]: ...
    def __hash__(self) -> builtins.int: ...
    def __new__(cls, frame:FrameIdentifier, frequency:Expression) -> ShiftFrequency: ...
    def __repr__(self) -> builtins.str: ...
    def to_quil(self) -> builtins.str: ...
    def to_quil_or_debug(self) -> builtins.str: ...

class ShiftPhase:
    @property
    def frame(self) -> FrameIdentifier: ...
    @property
    def phase(self) -> Expression: ...
    def __eq__(self, other:builtins.object) -> builtins.bool: ...
    def __getnewargs__(self) -> tuple[FrameIdentifier, Expression]: ...
    def __hash__(self) -> builtins.int: ...
    def __new__(cls, frame:FrameIdentifier, phase:Expression) -> ShiftPhase: ...
    def __repr__(self) -> builtins.str: ...
    def to_quil(self) -> builtins.str: ...
    def to_quil_or_debug(self) -> builtins.str: ...

class Store:
    @property
    def destination(self) -> builtins.str: ...
    @property
    def offset(self) -> MemoryReference: ...
    @property
    def source(self) -> ArithmeticOperand: ...
    def __eq__(self, other:builtins.object) -> builtins.bool: ...
    def __getnewargs__(self) -> tuple[builtins.str, MemoryReference, ArithmeticOperand]: ...
    def __hash__(self) -> builtins.int: ...
    def __new__(cls, destination:builtins.str, offset:MemoryReference, source:ArithmeticOperand) -> Store: ...
    def __repr__(self) -> builtins.str: ...
    def to_quil(self) -> builtins.str: ...
    def to_quil_or_debug(self) -> builtins.str: ...

class SwapPhases:
    @property
    def frame_1(self) -> FrameIdentifier: ...
    @property
    def frame_2(self) -> FrameIdentifier: ...
    def __eq__(self, other:builtins.object) -> builtins.bool: ...
    def __getnewargs__(self) -> tuple[FrameIdentifier, FrameIdentifier]: ...
    def __hash__(self) -> builtins.int: ...
    def __new__(cls, frame_1:FrameIdentifier, frame_2:FrameIdentifier) -> SwapPhases: ...
    def __repr__(self) -> builtins.str: ...
    def to_quil(self) -> builtins.str: ...
    def to_quil_or_debug(self) -> builtins.str: ...

class Target:
    def __getnewargs__(self) -> tuple[str | TargetPlaceholder]: ...
    def __repr__(self) -> builtins.str: ...
    def to_quil(self) -> builtins.str: ...
    def to_quil_or_debug(self) -> builtins.str: ...
    class Fixed(Target):
        __match_args__ = ("_0",)
        @property
        def _0(self) -> builtins.str: ...
        def __getitem__(self, key:builtins.int) -> typing.Any: ...
        def __len__(self) -> builtins.int: ...
        def __new__(cls, _0:builtins.str) -> Target.Fixed: ...
    
    class Placeholder(Target):
        __match_args__ = ("_0",)
        @property
        def _0(self) -> TargetPlaceholder: ...
        def __getitem__(self, key:builtins.int) -> typing.Any: ...
        def __len__(self) -> builtins.int: ...
        def __new__(cls, _0:TargetPlaceholder) -> Target.Placeholder: ...
    

class TargetPlaceholder:
    r"""
    An opaque placeholder for a label whose index may be assigned
    at a later time.
    """
    @property
    def base_label(self) -> builtins.str: ...
    def __eq__(self, other:builtins.object) -> builtins.bool: ...
    def __ge__(self, other:builtins.object) -> builtins.bool: ...
    def __gt__(self, other:builtins.object) -> builtins.bool: ...
    def __hash__(self) -> builtins.int: ...
    def __le__(self, other:builtins.object) -> builtins.bool: ...
    def __lt__(self, other:builtins.object) -> builtins.bool: ...
    def __new__(cls, base_label:builtins.str) -> TargetPlaceholder: ...
    def __repr__(self) -> builtins.str: ...

class UnaryLogic:
    @property
    def operand(self) -> MemoryReference: ...
    @property
    def operator(self) -> UnaryOperator: ...
    def __eq__(self, other:builtins.object) -> builtins.bool: ...
    def __getnewargs__(self) -> tuple[UnaryOperator, MemoryReference]: ...
    def __hash__(self) -> builtins.int: ...
    def __new__(cls, operator:UnaryOperator, operand:MemoryReference) -> UnaryLogic: ...
    def __repr__(self) -> builtins.str: ...
    def to_quil(self) -> builtins.str: ...
    def to_quil_or_debug(self) -> builtins.str: ...

class Vector:
    @property
    def data_type(self) -> ScalarType: ...
    @property
    def length(self) -> builtins.int: ...
    def __eq__(self, other:builtins.object) -> builtins.bool: ...
    def __getnewargs__(self) -> tuple[ScalarType, builtins.int]: ...
    def __hash__(self) -> builtins.int: ...
    def __new__(cls, data_type:ScalarType, length:builtins.int) -> Vector: ...
    def __repr__(self) -> builtins.str: ...
    def to_quil(self) -> builtins.str: ...
    def to_quil_or_debug(self) -> builtins.str: ...

class Waveform:
    @property
    def matrix(self) -> builtins.list[Expression]: ...
    @property
    def parameters(self) -> builtins.list[builtins.str]: ...
    def __eq__(self, other:builtins.object) -> builtins.bool: ...
    def __hash__(self) -> builtins.int: ...
    def __new__(cls, matrix:typing.Sequence[Expression], parameters:typing.Sequence[builtins.str]) -> Waveform: ...
    def __repr__(self) -> builtins.str: ...

class WaveformDefinition:
    @property
    def definition(self) -> Waveform: ...
    @property
    def name(self) -> builtins.str: ...
    def __eq__(self, other:builtins.object) -> builtins.bool: ...
    def __getnewargs__(self) -> tuple[builtins.str, Waveform]: ...
    def __hash__(self) -> builtins.int: ...
    def __new__(cls, name:builtins.str, definition:Waveform) -> WaveformDefinition: ...
    def __repr__(self) -> builtins.str: ...
    def to_quil(self) -> builtins.str: ...
    def to_quil_or_debug(self) -> builtins.str: ...

class WaveformInvocation:
    @property
    def name(self) -> builtins.str: ...
    @property
    def parameters(self) -> builtins.dict[builtins.str, Expression]: ...
    def __eq__(self, other:builtins.object) -> builtins.bool: ...
    def __getnewargs__(self) -> tuple[builtins.str, builtins.dict[builtins.str, Expression]]: ...
    def __new__(cls, name:builtins.str, parameters:typing.Mapping[builtins.str, Expression]) -> WaveformInvocation: ...
    def __repr__(self) -> builtins.str: ...
    def to_quil(self) -> builtins.str: ...
    def to_quil_or_debug(self) -> builtins.str: ...

class ArithmeticOperator(Enum):
    Add = ...
    Subtract = ...
    Divide = ...
    Multiply = ...

    def __repr__(self) -> builtins.str: ...
    def to_quil(self) -> builtins.str: ...
    def to_quil_or_debug(self) -> builtins.str: ...

class BinaryOperand(Enum):
    LiteralInteger = ...
    MemoryReference = ...

    def __repr__(self) -> builtins.str: ...
    def to_quil(self) -> builtins.str: ...
    def to_quil_or_debug(self) -> builtins.str: ...

class BinaryOperator(Enum):
    And = ...
    Ior = ...
    Xor = ...

    def __repr__(self) -> builtins.str: ...
    def to_quil(self) -> builtins.str: ...
    def to_quil_or_debug(self) -> builtins.str: ...

class ComparisonOperator(Enum):
    Equal = ...
    GreaterThanOrEqual = ...
    GreaterThan = ...
    LessThanOrEqual = ...
    LessThan = ...

    def __repr__(self) -> builtins.str: ...
    def to_quil(self) -> builtins.str: ...
    def to_quil_or_debug(self) -> builtins.str: ...

class GateModifier(Enum):
    r"""
    An enum of all the possible modifiers on a quil [`Gate`]
    """
    Controlled = ...
    r"""
    The `CONTROLLED` modifier makes the gate take an extra [`Qubit`] parameter as a control
    qubit.
    """
    Dagger = ...
    r"""
    The `DAGGER` modifier does a complex-conjugate transpose on the [`Gate`].
    """
    Forked = ...
    r"""
    The `FORKED` modifier allows an alternate set of parameters to be used based on the state
    of a qubit.
    """

    def __repr__(self) -> builtins.str: ...
    def to_quil(self) -> builtins.str: ...
    def to_quil_or_debug(self) -> builtins.str: ...

class PauliGate(Enum):
    I = ...
    X = ...
    Y = ...
    Z = ...

    def __repr__(self) -> builtins.str: ...
    @staticmethod
    def parse(input:builtins.str) -> PauliGate:
        r"""
        Parse a ``PauliGate`` from a string.
        
        Raises a ``ParseExpressionError`` error if the string isn't a valid Quil expression.
        """

class ScalarType(Enum):
    Bit = ...
    Integer = ...
    Octet = ...
    Real = ...

    def __repr__(self) -> builtins.str: ...
    def to_quil(self) -> builtins.str: ...
    def to_quil_or_debug(self) -> builtins.str: ...

class UnaryOperator(Enum):
    Neg = ...
    Not = ...

    def __repr__(self) -> builtins.str: ...
    def to_quil(self) -> builtins.str: ...
    def to_quil_or_debug(self) -> builtins.str: ...




import pickle
from quil.instructions import Instruction, Gate, MemoryReference, Qubit

class TestPickle:
    def test_gate(self):
        q0 = Qubit.Fixed(0)
        g = Gate("H", [], [q0], [])

        p = pickle.dumps(g)
        g2 = pickle.loads(p)
        assert g == g2

        i = Instruction.Gate(g)
        p2 = pickle.dumps(i)
        i2 = pickle.loads(p2)
        assert i == i2

    def test_halt(self):
        h = Instruction.Halt()
        p = pickle.dumps(h)
        q = pickle.loads(p)
        assert q == h

class TestParse:
    def test_memory_ref(self):
        input_str = "ro[0]"
        ref = MemoryReference.parse(input_str)
        assert isinstance(ref, MemoryReference)
        assert ref.index == 0
        assert ref.name == "ro"
        assert ref.to_quil() == input_str



import math
import random
from random import Random

from pytest import fixture


@fixture
def rng(seed: int) -> Random:
    """Get an RNG for tests where the actual parameters are somewhat arbitrary,
    and instead you're testing specific behavior.

    By default, a preset seed is used each run.
    You can set that seed directly as with the following example::

        pytest --seed 62590

    Alternatively, you can run a number of random trials,
    each getting a different randomized seed.
    This run 100 tests with different random seeds::

        pytest --trials 100

    You can combine these options to test first with a specific seed,
    then with a number of random seeds.
    """
    return Random(seed)


def pytest_addoption(parser):
    """Add new options for pytest tests."""

    parser.addoption("--trials", help="set a number of random test runs to perform")
    parser.addoption("--seed", help="set a specific seed to use")


def pytest_generate_tests(metafunc):
    if "seed" in metafunc.fixturenames:
        given_seed = metafunc.config.getoption("seed")
        seeds = [int(given_seed or 62590)]

        # Create random seeds if additional trials requested.
        trials = max(0, int(metafunc.config.getoption("trials") or 0))
        if trials <= 0 and not given_seed:
            trials = 1
        seeds.extend(random.randint(0, (1 << 32) - 1) for _ in range(trials))

        metafunc.parametrize("seed", seeds)


@fixture
def phase(rng: Random) -> float:
    r"""Return a random phase between :math:`0` and :math:`2\pi`."""

    return rng.uniform(0, math.pi * 2)


@fixture
def sample_count(rng: Random) -> int:
    r"""Return a random sample count."""

    return rng.randint(0, (1 << 64) - 1)


@fixture
def duration(rng: Random) -> float:
    r"""Return a random duration."""

    return rng.uniform(0, 1e10)


@fixture
def risetime(rng: Random) -> float:
    r"""Return a random risetime."""

    return rng.uniform(0, 1e10)


@fixture
def sample_rate(rng: Random) -> float:
    r"""Return a random sample_rate."""

    return rng.uniform(0, 1e10)


@fixture
def second_order_hrm_coeff(rng: Random) -> float:
    r"""Return a random second_order_hrm_coeff."""

    return rng.uniform(0, 1e10)


@fixture
def alpha(rng: Random) -> float:
    r"""Return a random alpha."""

    return rng.uniform(0, 1e10)


@fixture
def anh(rng: Random) -> float:
    r"""Return a random anharmonicity constant."""

    return rng.uniform(0, 1e10)


@fixture
def fwhm(rng: Random) -> float:
    r"""Return a random full width at half maximum value."""

    return rng.uniform(0, 1e10)


@fixture
def t0(rng: Random) -> float:
    r"""Return a random t0."""

    return rng.uniform(0, 1e10)


@fixture
def pad_left(rng: Random) -> float:
    r"""Return a random pad_left."""

    return rng.uniform(0, 1e10)


@fixture
def pad_right(rng: Random) -> float:
    r"""Return a random pad_right."""

    return rng.uniform(0, 1e10)


@fixture
def positive_polarity(rng: Random) -> bool:
    r"""Return a random positive polarity."""

    return bool(rng.randint(0, 1))


@fixture
def scale(rng: Random) -> float:
    r"""Return a random scale."""

    return rng.uniform(0, 1e10)


@fixture
def detuning(rng: Random) -> float:
    r"""Return a random detuning."""

    return rng.uniform(0, 1e10)



# This file is automatically generated by pyo3_stub_gen
# ruff: noqa: E501, F401

import builtins

class IdentifierValidationError(builtins.QuilError):
    r"""
    Errors that may occur when validating a Quil identifier.
    """
    ...

def validate_identifier(ident:builtins.str) -> None:
    r"""
    Returns an error if the given identifier is not a valid Quil Identifier.
    """

def validate_user_identifier(ident:builtins.str) -> None:
    r"""
    Returns an error if the given identifier is reserved, or if it is not a valid Quil identifier
    """




import copy
from typing import Any, Callable, Generator, Iterator

import pytest

import quil
from quil.expression import Expression
from quil.instructions import (
    CalibrationDefinition,
    CalibrationIdentifier,
    Delay,
    FrameIdentifier,
    Instruction,
    Pulse,
    Qubit,
    QubitPlaceholder,
    WaveformInvocation,
)


@pytest.fixture(params=(Qubit.Placeholder(QubitPlaceholder()), Qubit.Fixed(1), Qubit.Variable("x")))
def qubit(request: pytest.FixtureRequest) -> Qubit:
    assert isinstance(request.param, Qubit)
    return request.param


def _calibration(qubit: Qubit) -> tuple[CalibrationDefinition, Instruction]:
    calibration = CalibrationDefinition(
        CalibrationIdentifier("MYCAL", [], [qubit], []),
        [Instruction.Delay(Delay(Expression.Number(complex(0.5)), [], [qubit]))],
    )
    return calibration, Instruction.CalibrationDefinition(calibration)


def _pulse(qubit: Qubit) -> tuple[Pulse, Instruction]:
    pulse = Pulse(blocking=False, frame=FrameIdentifier("frame", [qubit]), waveform=WaveformInvocation("wf", {}))
    instr = Instruction.Pulse(pulse)
    return (pulse, instr)


@pytest.fixture(params=(_calibration, _pulse))
def make_instr(request: pytest.FixtureRequest) -> Iterator[Callable[[Qubit], tuple[Any, Instruction]]]:
    return request.param


def test_copy(qubit: Qubit, make_instr: Callable[[Qubit], tuple[Any, Instruction]]):
    inner, instr = make_instr(qubit)

    assert copy.copy(inner) == inner
    assert copy.copy(instr) == instr

    if not isinstance(qubit, Qubit.Placeholder):
        assert copy.deepcopy(inner) == inner
        assert copy.deepcopy(instr) == instr
    else:
        with pytest.raises(quil.PickleError, match=r"\bQubitPlaceholder\b"):
            _ = copy.deepcopy(inner)
        with pytest.raises(quil.PickleError, match=r"\bQubitPlaceholder\b"):
            _ = copy.deepcopy(instr)




"""The `quil` package provides tools for constructing, manipulating, parsing, and printing [Quil](https://github.com/quil-lang/quil) programs.

⚠️ This package is still in early development and breaking changes should be expected between minor versions.
"""

# The following code exposes the _quil package contents under the namespace quil.
from . import _quil
assert isinstance(_quil.__all__, list) and all(isinstance(s, str) for s in _quil.__all__)
exec(f"from ._quil import {', '.join(_quil.__all__)}; __all__ = {_quil.__all__}")
del _quil
